# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""This module attaches the HDF5 input output methods to the DataQualityFlag.

While these methods are avialable as methods of the class itself,
this module attaches them to the unified I/O registry, making it a bit
cleaner.
"""

import h5py
import numpy
from distutils.version import LooseVersion

from astropy.io.registry import (register_reader, register_writer,
                                 register_identifier)
from astropy.units import (UnitBase, Quantity)

from glue.lal import LIGOTimeGPS

from ... import version
from ...io.hdf5 import (open_hdf5, identify_hdf5)
from ..flag import DataQualityFlag
from ..segments import (SegmentList, Segment)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


def flag_from_hdf5(f, name=None, gpstype=LIGOTimeGPS, coalesce=True, nproc=1):
    """Read a `DataQualityFlag` object from an HDF5 file or group.
    """
    # hook multiprocessing
    if nproc != 1:
        return DataQualityFlag.read(f, name, coalesce=coalesce,
                                    gpstype=gpstype, format='cache')

    # open file is needed
    h5file = open_hdf5(f)

    try:
        # find dataset
        if name:
            dqfgroup = h5file[name]
        elif isinstance(h5file, h5py.File) and len(h5file) == 1:
            dqfgroup = h5file.values()[0]
        else:
            dqfgroup = h5file

        active = SegmentList.read(dqfgroup['active'])
        try:
            known = SegmentList.read(dqfgroup['known'])
        except KeyError as e:
            try:
                known = SegmentList.read(dqfgroup['valid'])
            except KeyError:
                raise e

        # read array, close file, and return
        out = DataQualityFlag(active=active, known=known,
                              **dict(dqfgroup.attrs))
    finally:
        if not isinstance(f, (h5py.Dataset, h5py.Group)):
            h5file.close()

    return out


def flag_to_hdf5(flag, output, name=None, group=None, compression='gzip',
                 **kwargs):
    """Write this `DataQualityFlag` to a :class:`h5py.Group`.

    This allows writing to an HDF5-format file.

    Parameters
    ----------
    output : `str`, :class:`h5py.Group`
        path to new output file, or open h5py `Group` to write to.
    name : `str`, optional
        custom name for this `Array` in the HDF hierarchy, defaults
        to the `name` attribute of the `Array`.
    group : `str`, optional
        parent group to create for this time-series.
    compression : `str`, optional
        name of compression filter to use
    **kwargs
        other keyword arguments passed to
        :meth:`h5py.Group.create_dataset`.

    Returns
    -------
    dqfgroup : :class:`h5py.Group`
        HDF group containing these data. This group contains 'active'
        and 'known' datasets, and metadata attrs.
    """
    # create output object
    import h5py
    if isinstance(output, h5py.Group):
        h5file = output
    else:
        h5file = h5py.File(output, 'w')

    try:
        # if group
        if group:
            try:
                h5group = h5file[group]
            except KeyError:
                h5group = h5file.create_group(group)
        else:
            h5group = h5file

        # create dataset
        name = name or flag.name
        if name is None:
            raise ValueError("Cannot store DataQualityFlag without a name. "
                             "Either assign the name attribute of the flag "
                             "itself, or given name= as a keyword argument to "
                             "write().")
        dqfgroup = h5group.create_group(name)

        flag.active.write(dqfgroup, 'active', compression=compression,
                          **kwargs)
        flag.known.write(dqfgroup, 'known', compression=compression,
                         **kwargs)

        # store metadata
        for attr in ['name', 'label', 'category', 'description', 'isgood',
                     'padding']:
            value = getattr(flag, attr)
            if value is None:
                continue
            elif isinstance(value, Quantity):
                dqfgroup.attrs[attr] = value.value
            elif isinstance(value, UnitBase):
                dqfgroup.attrs[attr] = str(value)
            else:
                dqfgroup.attrs[attr] = value

    finally:
        if not isinstance(output, (h5py.Dataset, h5py.Group)):
            h5file.close()

    return dqfgroup


def segmentlist_from_hdf5(f, name=None, gpstype=LIGOTimeGPS):
    """Read a `SegmentList` object from an HDF5 file or group.
    """
    h5file = open_hdf5(f)

    try:
        # find dataset
        if isinstance(h5file, h5py.Dataset):
            dataset = h5file
        else:
            dataset = h5file[name]

        try:
            data = dataset[()]
        except ValueError:
            data = []

        out = SegmentList()
        for row in data:
            row = map(int, row)
            # extract as LIGOTimeGPS
            start = LIGOTimeGPS(*row[:2])
            end = LIGOTimeGPS(*row[2:])
            # convert to user type
            try:
                start = gpstype(start)
            except TypeError:
                start = gpstype(float(start))
            try:
                end = gpstype(end)
            except TypeError:
                end = gpstype(float(end))
            out.append(Segment(start, end))
    finally:
        if not isinstance(f, (h5py.Dataset, h5py.Group)):
            h5file.close()

    return out


def segmentlist_to_hdf5(seglist, output, name, group=None,
                        compression='gzip', **kwargs):
    """Write a `SegmentList`
    """
    # create output object
    import h5py
    if isinstance(output, h5py.Group):
        h5file = output
    else:
        h5file = h5py.File(output, 'w')

    try:
        # if group
        if group:
            try:
                h5group = h5file[group]
            except KeyError:
                h5group = h5file.create_group(group)
        else:
            h5group = h5file

        # create dataset
        data = numpy.zeros((len(seglist), 4), dtype=int)
        for i, seg in enumerate(seglist):
            start, end = map(LIGOTimeGPS, seg)
            data[i, :] = (start.seconds, start.nanoseconds,
                          end.seconds, end.nanoseconds)
        if (not len(seglist) and
                LooseVersion(h5py.version.version).version[0] < 2):
            kwargs.setdefault('maxshape', (None, 4))
            kwargs.setdefault('chunks', (1, 1))
        dset = h5group.create_dataset(name, data=data,
                                      compression=compression, **kwargs)
    finally:
        if not isinstance(output, h5py.Group):
            h5file.close()

    return dset


# register HDF5 format
register_reader('hdf5', SegmentList, segmentlist_from_hdf5)
register_writer('hdf5', SegmentList, segmentlist_to_hdf5)
register_identifier('hdf5', SegmentList, identify_hdf5)
register_reader('hdf5', DataQualityFlag, flag_from_hdf5)
register_writer('hdf5', DataQualityFlag, flag_to_hdf5)
register_identifier('hdf5', DataQualityFlag, identify_hdf5)

# register generic HDF format to HDF5
register_reader('hdf', SegmentList, segmentlist_from_hdf5)
register_writer('hdf', SegmentList, segmentlist_to_hdf5)
register_reader('hdf', DataQualityFlag, flag_from_hdf5)
register_writer('hdf', DataQualityFlag, flag_to_hdf5)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilies for interacting with the LIGO Algorithm Library.
"""

from __future__ import absolute_import

from six import string_types

import numpy

from astropy import units

from lal import lal

from .. import version
from ..time import to_gps

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version

# LAL type enum
try:
    LAL_TYPE_STR = {lal.LAL_I2_TYPE_CODE: 'INT2',
                    lal.LAL_I4_TYPE_CODE: 'INT4',
                    lal.LAL_I8_TYPE_CODE: 'INT8',
                    lal.LAL_U2_TYPE_CODE: 'UINT2',
                    lal.LAL_U4_TYPE_CODE: 'UINT4',
                    lal.LAL_U8_TYPE_CODE: 'UINT8',
                    lal.LAL_S_TYPE_CODE: 'REAL4',
                    lal.LAL_D_TYPE_CODE: 'REAL8',
                    lal.LAL_C_TYPE_CODE: 'COMPLEX8',
                    lal.LAL_Z_TYPE_CODE: 'COMPLEX16'}
except AttributeError:
    LAL_TYPE_STR = {lal.I2_TYPE_CODE: 'INT2',
                    lal.I4_TYPE_CODE: 'INT4',
                    lal.I8_TYPE_CODE: 'INT8',
                    lal.U2_TYPE_CODE: 'UINT2',
                    lal.U4_TYPE_CODE: 'UINT4',
                    lal.U8_TYPE_CODE: 'UINT8',
                    lal.S_TYPE_CODE: 'REAL4',
                    lal.D_TYPE_CODE: 'REAL8',
                    lal.C_TYPE_CODE: 'COMPLEX8',
                    lal.Z_TYPE_CODE: 'COMPLEX16'}

LAL_TYPE_FROM_STR = dict((v, k) for k, v in LAL_TYPE_STR.iteritems())

# map numpy dtypes to LAL type codes
try:
    LAL_TYPE_FROM_NUMPY = {numpy.int16: lal.LAL_I2_TYPE_CODE,
                           numpy.int32: lal.LAL_I4_TYPE_CODE,
                           numpy.int64: lal.LAL_I8_TYPE_CODE,
                           numpy.uint16: lal.LAL_U2_TYPE_CODE,
                           numpy.uint32: lal.LAL_U4_TYPE_CODE,
                           numpy.uint64: lal.LAL_U8_TYPE_CODE,
                           numpy.float32: lal.LAL_S_TYPE_CODE,
                           numpy.float64: lal.LAL_D_TYPE_CODE,
                           numpy.complex64: lal.LAL_C_TYPE_CODE,
                           numpy.complex128: lal.LAL_Z_TYPE_CODE}
except AttributeError:
    LAL_TYPE_FROM_NUMPY = {numpy.int16: lal.I2_TYPE_CODE,
                           numpy.int32: lal.I4_TYPE_CODE,
                           numpy.int64: lal.I8_TYPE_CODE,
                           numpy.uint16: lal.U2_TYPE_CODE,
                           numpy.uint32: lal.U4_TYPE_CODE,
                           numpy.uint64: lal.U8_TYPE_CODE,
                           numpy.float32: lal.S_TYPE_CODE,
                           numpy.float64: lal.D_TYPE_CODE,
                           numpy.complex64: lal.C_TYPE_CODE,
                           numpy.complex128: lal.Z_TYPE_CODE}

LAL_TYPE_STR_FROM_NUMPY = dict((key, LAL_TYPE_STR[value]) for (key, value) in
                               LAL_TYPE_FROM_NUMPY.iteritems())

try:
    LAL_UNIT_INDEX = [lal.lalMeterUnit,
                      lal.lalKiloGramUnit,
                      lal.lalSecondUnit,
                      lal.lalAmpereUnit,
                      lal.lalKelvinUnit,
                      lal.lalStrainUnit,
                      lal.lalADCCountUnit]
except AttributeError:
    LAL_UNIT_INDEX = [lal.MeterUnit,
                      lal.KiloGramUnit,
                      lal.SecondUnit,
                      lal.AmpereUnit,
                      lal.KelvinUnit,
                      lal.StrainUnit,
                      lal.ADCCountUnit]
    lal_unit_to_str = str
    LAL_UNIT_FROM_ASTROPY = dict((units.Unit(lal_unit_to_str(u)), u) for
                                 u in LAL_UNIT_INDEX)
else:
    lal_unit_to_str = lal.UnitToString
    LAL_UNIT_FROM_ASTROPY = dict((units.Unit(lal_unit_to_str(u)), u) for
                                 u in LAL_UNIT_INDEX)


def to_lal_unit(aunit):
    """Convert the input unit into a :lal:`LALUnit`

    For example::

       >>> u = to_lal_unit('m**2 / kg ** 4')
       >>> print(u)
       m^2 kg^-4

    Parameters
    ----------
    aunit : `~astropy.units.Unit`, `str`
        the input unit

    Returns
    -------
    unit : :lal:`LALUnit`
        the LALUnit representation of the input

    Raises
    ------
    ValueError
        if LAL doesn't understand the base units for the input
    """
    if isinstance(aunit, string_types):
        aunit = units.Unit(aunit)
    lunit = lal.Unit()
    for base, power in zip(aunit.bases, aunit.powers):
        # try this base
        try:
            lalbase = LAL_UNIT_FROM_ASTROPY[base]
        except KeyError:
            lalbase = None
            # otherwise loop through the equivalent bases
            for eqbase in base.find_equivalent_units():
                try:
                    lalbase = LAL_UNIT_FROM_ASTROPY[eqbase]
                except KeyError:
                    continue
        # if we didn't find anything, raise an exception
        if lalbase is None:
            raise ValueError("LAL has no unit corresponding to %r" % base)
        lunit *= lalbase ** power
    return lunit


def from_lal_unit(lunit):
    """Convert a :lal`LALUnit` into a `~astropy.units.Unit`

    Parameters
    ----------
    aunit : :lal:`LALUnit`
        the input unit

    Returns
    -------
    unit : `~astropy.units.Unit`
        the Astropy representation of the input

    Raises
    ------
    ValueError
        if Astropy doesn't understand the base units for the input
    """
    aunit = None
    for power, lalbase in zip(lunit.unitNumerator, LAL_UNIT_INDEX):
        # if not used, continue
        if not power:
            continue
        # convert to astropy unit
        try:
            u = units.Unit(lal_unit_to_str(lalbase))
        except ValueError:
            raise ValueError("Astropy has no unit corresponding to %r"
                             % lalbase)
        if aunit is None:
            aunit = u ** power
        else:
            aunit *= u ** power
    return aunit


def to_lal_ligotimegps(gps):
    """Convert the given GPS time to a :lal:`LIGOTimeGPS` object

    Parameters
    ----------
    gps : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        input GPS time, can be anything parsable by :meth:`~gwpy.time.to_gps`

    Returns
    -------
    ligotimegps : :lal:`LIGOTimeGPS`
        a SWIG-LAL `LIGOTimeGPS` representation of the given GPS time
    """
    gps = to_gps(gps)
    return lal.LIGOTimeGPS(gps.seconds, gps.nanoseconds)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Input/output methods for tabular data.
"""

from glue.ligolw.ligolw import LIGOLWContentHandler

from .. import lsctables
lsctables.use_in(LIGOLWContentHandler)

from ... import version

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version

# import LIGO_LW I/O
from .ligolw import *

# try importing ROOT-based I/O
from .omicron import *

# import omega I/O
from .omega import *

# import cWB I/O
from .cwb import *

# import cache I/O
from .cache import *



from gwpy.timeseries import TimeSeries
gwdata = TimeSeries.fetch('H1:LDAS-STRAIN', 'September 16 2010 06:40', 'September 16 2010 06:50')
spectrum = gwdata.asd(8, 4)
plot = spectrum.plot()
ax = plot.gca()
ax.set_xlim(40, 4000)
ax.set_ylabel(r'Gravitational-wave strain ASD [strain$/\sqrt{\mathrm{Hz}}$]')
ax.set_ylim(1e-23, 1e-19)
plot.show()


# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read gravitational-wave frame (GWF) files using the LALFrame API.
"""

from __future__ import division

from ....io.cache import (CacheEntry, file_list)
from ....time import to_gps
from ....utils import (import_method_dependency, with_import)
from ... import (TimeSeries, TimeSeriesDict)
from . import channel_dict_kwarg

DEPENDS = 'lalframe.frread'


@with_import(DEPENDS)
def read_timeseriesdict(source, channels, start=None, end=None, dtype=None,
                        resample=None, verbose=False, _SeriesClass=TimeSeries):
    """Read the data for a list of channels from a GWF data source.

    Parameters
    ----------
    source : `str`, :class:`glue.lal.Cache`, `list`
        data source object, one of:

        - `str` : frame file path
        - :class:`glue.lal.Cache`, `list` : contiguous list of frame paths

    channels : `list`
        list of channel names (or `Channel` objects) to read from frame.
    start : `Time`, :lalsuite:`LIGOTimeGPS`, optional
        start GPS time of desired data.
    end : `Time`, :lalsuite:`LIGOTimeGPS`, optional
        end GPS time of desired data.
    type : `str`
        type of channel, one of ``'adc'`` or ``'proc'``.
    dtype : `numpy.dtype`, `str`, `type`, or `dict`
        numeric data type for returned data, e.g. `numpy.float`, or
        `dict` of (`channel`, `dtype`) pairs
    resample : `float`, optional
        rate of samples per second at which to resample input TimeSeries.
    verbose : `bool`, optional
        print verbose output.

    Returns
    -------
    dict : :class:`~gwpy.timeseries.TimeSeriesDict`
        dict of (channel, `TimeSeries`) data pairs

    Notes
    -----
    If reading from a list, or cache, or framefiles, the frames contained
    must be contiguous and sorted in chronological order for this function
    to return without exception.

    Raises
    ------
    ValueError
        if reading from an unsorted, or discontiguous cache of files
    """
    lal = import_method_dependency('lal.lal')
    from gwpy.utils.lal import to_lal_ligotimegps
    # parse input arguments (astropy insists on opening files)
    if isinstance(source, file):
        source = source.name
    filelist = file_list(source)
    try:
        frametype = CacheEntry.from_T050017(filelist[0]).description
    except ValueError:
        frametype = None
    # set times
    if start is not None:
        start = to_lal_ligotimegps(start)
    if end is not None:
        end = to_lal_ligotimegps(end)
    if start and end:
        duration = float(end - start)
    elif end:
        raise ValueError("If `end` is given, `start` must also be given")
    else:
        duration = None
    # parse resampling
    resample = channel_dict_kwarg(resample, channels, (int,))
    if resample is None:
        raise ValueError("Cannot parse resample request, please review "
                         "documentation for that argument")
    # parse dtype
    dtype = channel_dict_kwarg(dtype, channels, (str, type))
    if dtype is None:
        raise ValueError("Cannot parse dtype request, please review "
                         "documentation for that argument")

    # read data
    try:
        laldata = frread.read_timeseries(source, map(str, channels),
                                         start=start, duration=duration,
                                         verbose=verbose)
    # if old version of lalframe.frread
    except TypeError:
        laldata = [frread.read_timeseries(source, str(c), start=start,
                                          duration=duration, verbose=verbose)
                   for c in channels]
    # convert to native objects and return
    out = TimeSeriesDict()
    for channel, lalts in zip(channels, laldata):
        ts = _SeriesClass.from_lal(lalts)
        ts.channel.frametype = frametype
        if channel in dtype:
            ts = ts.astype(dtype[channel])
        if channel in resample:
            ts = ts.resample(resample[channel])
        out[channel] = ts
    return out



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Write GWpy objects to HDF5 files
"""

try:
    import h5py
    HAVE_H5PY = True
except ImportError:
    HAVE_H5PY = False

from .. import version
from ..utils.deps import with_import

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


@with_import('h5py')
def open_hdf5(filename):
    """Wrapper to open a :class:`h5py.File` from disk, gracefully
    handling a few corner cases
    """
    if isinstance(filename, (h5py.Group, h5py.Dataset)):
        return filename
    elif isinstance(filename, file):
        return h5py.File(filename.name, 'r')
    else:
        return h5py.File(filename, 'r')


def identify_hdf5(*args, **kwargs):
    """Identify an input file as LOSC HDF based on its filename
    """
    filename = args[3]
    if isinstance(filename, file):
        filename = filename.name
    if (isinstance(filename, (unicode, str)) and
            filename.endswith(('hdf', 'hdf5'))):
        return True
    elif HAVE_H5PY and isinstance(filename, (h5py.Group, h5py.Dataset)):
        return True
    else:
        return False



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Basic utilities for reading/writing LIGO_LW-format XML files.

All specific unified input/output for class objecst should be placed in
an 'io' subdirectory of the containing directory for that class.
"""

from .utils import GzipFile

from glue.lal import CacheEntry
from glue.ligolw.ligolw import (Document, LIGOLWContentHandler,
                                PartialLIGOLWContentHandler)
from glue.ligolw.table import CompareTableNames as compare_table_names
from glue.ligolw.utils.ligolw_add import ligolw_add
from glue.ligolw import (table, lsctables)

from .. import version
from ..utils import gprint
from .cache import file_list
from .utils import identify_factory

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


class GWpyContentHandler(LIGOLWContentHandler):
    """Empty sub-class of `~glue.ligolw.ligolw.LIGOLWContentHandler`
    """
    pass


def get_partial_contenthandler(table):
    """Build a `~glue.ligolw.ligolw.PartialLIGOLWContentHandler` for this table

    Parameters
    ----------
    table : `type`
        the table class to be read

    Returns
    -------
    contenthandler : `type`
        a subclass of `~glue.ligolw.ligolw.PartialLIGOLWContentHandler` to
        read only the given `table`
    """
    def _filter_func(name, attrs):
        if name == table.tagName and attrs.has_key('Name'):
            return compare_table_names(attrs.get('Name'), table.tableName) == 0
        else:
            return False

    class _ContentHandler(PartialLIGOLWContentHandler):
        def __init__(self, document):
            super(_ContentHandler, self).__init__(document, _filter_func)

    return _ContentHandler


def table_from_file(f, tablename, columns=None, filt=None,
                    contenthandler=None, nproc=1, verbose=False):
    """Read a `~glue.ligolw.table.Table` from a LIGO_LW file.

    Parameters
    ----------
    f : `file`, `str`, `CacheEntry`, `list`, `Cache`
        object representing one or more files. One of

        - an open `file`
        - a `str` pointing to a file path on disk
        - a formatted `~glue.lal.CacheEntry` representing one file
        - a `list` of `str` file paths
        - a formatted `~glue.lal.Cache` representing many files

    tablename : `str`
        name of the table to read.
    columns : `list`, optional
        list of column name strings to read, default all.
    filt : `function`, optional
        function by which to `filter` events. The callable must accept as
        input a row of the table event and return `True`/`False`.
    contenthandler : `~glue.ligolw.ligolw.LIGOLWContentHandler`
        SAX content handler for parsing LIGO_LW documents.

    Returns
    -------
    table : `~glue.ligolw.table.Table`
        `Table` of data with given columns filled
    """
    # find table class
    tableclass = lsctables.TableByName[table.StripTableName(tablename)]

    # get content handler
    if contenthandler is None:
        contenthandler = get_partial_contenthandler(tableclass)

    # allow cache multiprocessing
    if nproc != 1:
        return tableclass.read(f, columns=columns,
                               contenthandler=contenthandler,
                               nproc=nproc, format='cache')

    # set columns to read
    if columns is not None:
        _oldcols = tableclass.loadcolumns
        tableclass.loadcolumns = columns

    # generate Document and populate
    files = [fp.name if isinstance(fp, (file, GzipFile)) else
             fp for fp in file_list(f)]
    xmldoc = Document()
    ligolw_add(xmldoc, files, non_lsc_tables_ok=True,
               contenthandler=contenthandler, verbose=verbose)

    # extract table
    try:
        out = tableclass.get_table(xmldoc)
    except ValueError:
        out = lsctables.New(tableclass, columns=columns)
    if verbose:
        gprint('%d rows found in %s table' % (len(out), out.tableName))

    if filt:
        if verbose:
            gprint('filtering rows ...', end=' ')
        try:
            out_ = out.copy()
        except AttributeError:
            out_ = table.new_from_template(out)
        out_.extend(filter(filt, out))
        out = out_
        if verbose:
            gprint('%d rows remaining\n' % len(out))
    if columns is not None:
        tableclass.loadcolumns = _oldcols
    return out


identify_ligolw = identify_factory('xml', 'xml.gz')



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read event tables from ASCII files.

This module only defines a function factory for reading lines of ASCII
into a particular LIGO_LW table object.

Each specific ASCII table format should define their own line parser
(that generates each row of the table) and pass it to the factory method.
"""

from numpy import loadtxt

from astropy.io.registry import (register_reader, register_identifier)

from glue.ligolw.table import (reassign_ids, StripTableName)

from ..lsctables import (New, TableByName)
from ..utils import TIME_COLUMN
from ... import version
from ...io.cache import file_list
from ...io.utils import identify_factory
from ...time import LIGOTimeGPS

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


def return_reassign_ids(elem):
    """Wrapper to `glue.ligolw.table.reassign_ids` that returns
    """
    reassign_ids(elem)
    return elem


def table_from_ascii_factory(table, format, trig_func, cols=None, **kwargs):
    """Build a table reader for the given format

    Parameters
    ----------
    table : `type`
        table class for which this format is relevant
    format : `str`
        name of the format
    trig_func : `callable`
        method to convert one row of data (from `numpy.loadtxt`) into an event
    cols : `list` of `str`
        list of columns that can be read by default for this format
    **kwargs
        default keyword arguments to pass to `numpy.loadtxt`

    Returns
    -------
    row_reader : `function`
        function that can be used to read a row of this table from ascii.
        The returned function natively supports multi-processing.
    """
    def table_from_ascii_rows(
            f, columns=cols, filt=None, nproc=1, **loadtxtkwargs):
        """Build a `~{0}` from events in an ASCII file.

        Parameters
        ----------
        f : `file`, `str`, `CacheEntry`, `list`, `Cache`
            object representing one or more files. One of

            - an open `file`
            - a `str` pointing to a file path on disk
            - a formatted :class:`~glue.lal.CacheEntry` representing one file
            - a `list` of `str` file paths
            - a formatted :class:`~glue.lal.Cache` representing many files

        columns : `list`, optional
            list of column name strings to read, default all.
        filt : `function`, optional
            function by which to filt events. The callable must accept as
            input a `SnglBurst` event and return `True`/`False`.
        nproc : `int`, optional, default: 1
            number of parallel processes with which to distribute file I/O,
            default: serial process
        **loadtxtkwargs
            all other keyword arguments are passed to `numpy.loadtxt`

        Returns
        -------
        table : `~{0}`
            a new `~{0}` filled with yummy data
        """.format(table.__name__)
        # format keyword arguments
        kwargs_ = kwargs.copy()
        kwargs_.update(loadtxtkwargs)

        # format list of files
        files = file_list(f)

        # allow multiprocessing
        if nproc != 1:
            from ...io.cache import read_cache
            return read_cache(files, table, nproc, return_reassign_ids,
                              columns=columns, format=format, **kwargs_)

        # work out columns to read from ASCII
        if columns is None:
            columns = cols
        else:
            columns = list(columns)
        # and translate them into LIGO_LW columns (only for 'time')
        try:
            ligolwcolumns = list(columns)
        except TypeError as e:
            e.args = ('This ascii format requires the column list to be given '
                      'manually, please give the `columns` keyword argument',)
            raise
        if 'time' in columns and table.tableName in TIME_COLUMN:
            ligolwcolumns.remove('time')
            ligolwcolumns.extend(TIME_COLUMN[table.tableName])

        # generate output
        out = New(table, columns=ligolwcolumns)
        append = out.append

        # iterate over events
        for fp in files:
            dat = loadtxt(fp, **kwargs_)
            for line in dat:
                row = trig_func(line, columns=columns)
                if 'event_id' in ligolwcolumns:
                    row.event_id = out.get_next_id()
                if filt is None or filt(row):
                    append(row)
        return out
    return table_from_ascii_rows


def row_from_ascii_factory(table, delimiter):
    """Build a generic ASCII table row reader for the given delimiter
    """
    name = table.tableName
    RowType = table.RowType
    tcols = TIME_COLUMN.get(name, None)

    def row_from_ascii(line, columns):
        """Build a `~{0}` from a line of ASCII data

        Parameters
        ----------
        line : `str`, `array-like`
            a line of ASCII data, either as a delimited string, or an array
        columns : `list` of `str`
            the names of each of the ASCII columns, give 'time' for 'standard'
            time columns for a given table (e.g. 'end_time', and 'end_time_ns'
            for `SnglInspiralTable`)
        delimiter : `str`, optional, default: any whitespace
            the ASCII delimiter, only needed if `line` is given as a `str`

        Returns
        -------
        {1}: `~{0}`
            a new `~{0}` filled with data

        Raises
        ------
        AttributeError
            if any column is not recognised
        """.format(RowType.__name__, StripTableName(name))
        row = table.RowType()
        if isinstance(line, str):
            line = line.rstrip('\n').split(delimiter)
        line = map(float, line)
        for datum, colname in zip(line, columns):
            if colname == 'time' and tcols is not None:
                t = LIGOTimeGPS(datum)
                setattr(row, tcols[0], t.seconds)
                setattr(row, tcols[1], t.nanoseconds)
            else:
                try:
                    getattr(row, 'set_%s' % colname)(datum)
                except AttributeError:
                    setattr(row, colname, datum)
        return row
    return row_from_ascii

# register generic ASCII parsing for all tables
for table in TableByName.itervalues():
    # register whitespace-delimited ASCII
    register_reader(
        'ascii', table, table_from_ascii_factory(
            table, 'ascii', row_from_ascii_factory(table, None)))
    register_identifier('ascii', table, identify_factory('txt', 'txt.gz'))
    # register csv
    register_reader(
        'csv', table, table_from_ascii_factory(
            table, 'csv', row_from_ascii_factory(table, ','), delimiter=','))
    register_identifier('csv', table, identify_factory('csv', 'csv.gz'))



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for the `gwpy.cli` module
"""

import os
import tempfile
import importlib

from compat import unittest

from gwpy import version

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


class CliTestMixin(object):
    PRODUCT_NAME = 'gwpy.cli.cliproduct.CliProduct'

    def setUp(self):
        self.PRODUCT_TYPE = self._import_product()

    def _import_product(self):
        modname, objname = self.PRODUCT_NAME.rsplit('.', 1)
        mod = importlib.import_module(modname)
        return getattr(mod, objname)

    def test_init(self):
        product = self.PRODUCT_TYPE()


class CliTimeSeriesTests(CliTestMixin, unittest.TestCase):
    PRODUCT_NAME = 'gwpy.cli.timeseries.TimeSeries'


class CliSpectrumTests(CliTestMixin, unittest.TestCase):
    PRODUCT_NAME = 'gwpy.cli.spectrum.Spectrum'


class CliSpectrogramTests(CliTestMixin, unittest.TestCase):
    PRODUCT_NAME = 'gwpy.cli.spectrogram.Spectrogram'


class CliCoherenceTests(CliTestMixin, unittest.TestCase):
    PRODUCT_NAME = 'gwpy.cli.coherence.Coherence'


class CliCoherencegramTests(CliTestMixin, unittest.TestCase):
    PRODUCT_NAME = 'gwpy.cli.coherencegram.Coherencegram'


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Extension of the :class:`~matplotlib.axes.Axes` class with
user-friendly attributes
"""

from six import string_types

from matplotlib.axes import Axes as _Axes
from matplotlib.artist import Artist
from matplotlib.projections import register_projection

from .decorators import auto_refresh
from . import (rcParams, tex, html)

from .. import version
__version__ = version.version
__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


class Axes(_Axes):
    """An extension of the core matplotlib :class:`~matplotlib.axes.Axes`.

    These custom `Axes` provide only some simpler attribute accessors.

    Notes
    -----
    A new set of `Axes` should be constructed via::

        >>> plot.add_subplots(111, projection='xxx')

    where plot is a :class:`~gwpy.plotter.Plot` figure, and ``'xxx'``
    is the name of the `Axes` you want to add.
    """
    projection = 'rectilinear'

    def __init__(self, *args, **kwargs):
        super(Axes, self).__init__(*args, **kwargs)
        self.xaxis.labelpad = 10
    __init__.__doc__ = _Axes.__init__.__doc__

    # -----------------------------------------------
    # text properties

    # x-axis label
    @property
    def xlabel(self):
        """Label for the x-axis

        :type: :class:`~matplotlib.text.Text`
        """
        return self.xaxis.label

    @xlabel.setter
    @auto_refresh
    def xlabel(self, text):
        if isinstance(text, string_types):
            self.set_xlabel(text)
        else:
            self.xaxis.label = text

    @xlabel.deleter
    @auto_refresh
    def xlabel(self):
        self.set_xlabel("")

    # y-axis label
    @property
    def ylabel(self):
        """Label for the y-axis

        :type: :class:`~matplotlib.text.Text`
        """
        return self.yaxis.label

    @ylabel.setter
    @auto_refresh
    def ylabel(self, text):
        if isinstance(text, string_types):
            self.set_ylabel(text)
        else:
            self.yaxis.label = text

    @ylabel.deleter
    @auto_refresh
    def ylabel(self):
        self.set_ylabel("")

    # -----------------------------------------------
    # limit properties

    @property
    def xlim(self):
        """Limits for the x-axis

        :type: `tuple`
        """
        return self.get_xlim()

    @xlim.setter
    @auto_refresh
    def xlim(self, limits):
        self.set_xlim(*limits)

    @xlim.deleter
    @auto_refresh
    def xlim(self):
        self.relim()
        self.autoscale_view(scalex=True, scaley=False)

    @property
    def ylim(self):
        """Limits for the y-axis

        :type: `tuple`
        """
        return self.get_ylim()

    @ylim.setter
    @auto_refresh
    def ylim(self, limits):
        self.set_ylim(*limits)

    @ylim.deleter
    def ylim(self):
        self.relim()
        self.autoscale_view(scalex=False, scaley=True)

    # -----------------------------------------------
    # scale properties

    @property
    def logx(self):
        """Display the x-axis with a logarithmic scale

        :type: `bool`
        """
        return self.get_xscale() == "log"

    @logx.setter
    @auto_refresh
    def logx(self, log):
        if log and not self.logx:
            self.set_xscale('log')
        elif self.logx and not log:
            self.set_xscale('linear')

    @property
    def logy(self):
        """Display the y-axis with a logarithmic scale

        :type: `bool`
        """
        return self.get_yscale() == "log"

    @logy.setter
    @auto_refresh
    def logy(self, log):
        if log and not self.logy:
            self.set_yscale('log')
        elif self.logy and not log:
            self.set_yscale('linear')

    # -------------------------------------------
    # Axes methods

    @auto_refresh
    def resize(self, pos, which='both'):
        """Set the axes position with::

            pos = [left, bottom, width, height]

        in relative 0,1 coords, or *pos* can be a
        :class:`~matplotlib.transforms.Bbox`

        There are two position variables: one which is ultimately
        used, but which may be modified by :meth:`apply_aspect`, and a
        second which is the starting point for :meth:`apply_aspect`.
        """
        return super(Axes, self).set_position(pos, which=which)

    @auto_refresh
    def add_label_unit(self, unit, axis='x'):
        label = getattr(self, 'get_%slabel' % axis)()
        if not label:
            label = unit.__doc__
        if rcParams.get("text.usetex", False):
            unitstr = tex.unit_to_latex(unit)
        else:
            unitstr = unit.to_string()
        set_ = getattr(self, 'set_%slabel' % axis)
        if label:
            set_("%s [%s]" % (label, unitstr))
        else:
            set_(unitstr)

    def legend(self, *args, **kwargs):
        # set kwargs
        alpha = kwargs.pop("alpha", 0.8)
        linewidth = kwargs.pop("linewidth", 8)

        # make legend
        legend = super(Axes, self).legend(*args, **kwargs)
        # find relevant axes
        if legend is not None:
            lframe = legend.get_frame()
            lframe.set_alpha(alpha)
            [l.set_linewidth(linewidth) for l in legend.get_lines()]
        return legend
    legend.__doc__ = _Axes.legend.__doc__

    def html_map(self, imagefile, data=None, **kwargs):
        """Create an HTML map for some data contained in these `Axes`

        Parameters
        ----------
        data : `~matplotlib.artist.Artist`, `~gwpy.data.Series`, `array-like`
            data to map, one of an `Artist` already drawn on these axes (
            via :meth:`plot` or :meth:`scatter`, for example) or a data set
        imagefile : `str`
            path to image file on disk for the containing `Figure`
        mapname : `str`, optional
            ID to connect <img> tag and <map> tags, default: ``'points'``. This
            should be unique if multiple maps are to be written to a single
            HTML file.
        shape : `str`, optional
            shape for <area> tag, default: ``'circle'``
        standalone : `bool`, optional
            wrap map HTML with required HTML5 header and footer tags,
            default: `True`
        title : `str`, optional
            title name for standalone HTML page
        jquery : `str`, optional
            URL of jquery script, defaults to googleapis.com URL

        Returns
        -------
        HTML : `str`
            string of HTML markup that defines the <img> and <map>
        """
        if data is None:
            artists = self.lines + self.collections + self.images
            if len(artists) != 1:
                raise ValueError("Cannot determine artist to map, %d found."
                                 % len(artists))
            data = artists[0]
        if isinstance(data, Artist):
            return html.map_artist(data, imagefile, **kwargs)
        else:
            return html.map_data(data, self, imagefile, **kwargs)

register_projection(Axes)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------------
#
# GWpy documentation build configuration file, created by
# sphinx-quickstart on Thu Nov  7 14:39:30 2013.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import inspect
import os.path

from matplotlib import use
use('agg')

import sphinx_bootstrap_theme

import gwpy
from gwpy import version as gwpy_version
from gwpy.plotter import (GWPY_PLOT_PARAMS)

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#sys.path.insert(0, os.path.abspath('.'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.pngmath',
    'sphinx.ext.autosummary',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.linkcode',
    'numpydoc',
    'sphinxcontrib.epydoc',
    'sphinxcontrib.doxylink',
    'matplotlib.sphinxext.plot_directive',
    'gwpy.utils.sphinx.autoclassapi',
    'gwpy.utils.sphinx.directives',
]

# customise autodoc
autoclass_content = 'class'
autodoc_default_flags = ['show-inheritance', 'members', 'inherited-members']

# Epydoc extension config for GLUE
# Can de-comment when GLUE is actually referenced in the documentation
epydoc_mapping = {
    'https://www.lsc-group.phys.uwm.edu/daswg/projects/glue/doc/':
    [r'glue(\.|$)'],
}

# doxylink
LALSUITE_DOCS = 'http://software.ligo.org/docs/lalsuite'
doxylink = {
    'lal': ('lal.tag', '%s/lal/' % LALSUITE_DOCS),
    'lalframe': ('lalframe.tag', '%s/lalframe/' % LALSUITE_DOCS),
}

# matplotlib plot directive
plot_rcparams = GWPY_PLOT_PARAMS
plot_apply_rcparams = True
plot_formats = ['png']

# fix numpydoc autosummary
numpydoc_show_class_members = False

# set autosummary
autosummary_generate = False

# configure inheritance diagram
inheritance_graph_attrs = dict(rankdir='TB')

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'GWpy'
copyright = u'2013, Duncan Macleod'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = gwpy_version.version
# The full version, including alpha/beta/rc tags.
release = gwpy_version.version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', '_templates', '_generated']

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = 'obj'

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'bootstrap'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    'source_link_position': None,
    'navbar_site_name': "Contents",
    'navbar_links': [("Examples", "examples/index")],
    'navbar_sidebarrel': False,
    'navbar_pagenav': False,
}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = sphinx_bootstrap_theme.get_html_theme_path()

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = 'gwpy_white_24.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = 'favicon.png'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {'**': ['localtoc.html', 'sourcelink.html', 'searchbox.html']}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'GWpydoc'


# Intersphinx
intersphinx_mapping = {
    'python': ('http://docs.python.org/', None),
    'numpy': ('http://docs.scipy.org/doc/numpy/', None),
    'scipy': ('http://docs.scipy.org/doc/scipy/reference/', None),
    'matplotlib': ('http://matplotlib.sourceforge.net/', None),
    'astropy': ('http://docs.astropy.org/en/stable/', None),
    'gwpy': ('http://gwpy.github.io/docs/stable/', None),
    'pycbc': ('https://ldas-jobs.ligo.caltech.edu/~cbc/docs/pycbc/',
              'pycbc.inv'),
}


# -----------------------------------------------------------------------------
# Source code links

def linkcode_resolve(domain, info):
    """Determine the URL corresponding to Python object

    This code is stolen with thanks from the scipy team.
    """
    if domain != 'py':
        return None

    modname = info['module']
    fullname = info['fullname']

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split('.'):
        try:
            obj = getattr(obj, part)
        except:
            return None
    # try and sneak past a decorator
    try:
        obj = obj.im_func.func_closure[0].cell_contents
    except (AttributeError, TypeError):
        pass

    try:
        fn = inspect.getsourcefile(obj)
    except:
        fn = None
    if not fn:
        try:
            fn = inspect.getsourcefile(sys.modules[obj.__module__])
        except:
            fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.findsource(obj)
    except:
        lineno = None

    if lineno:
        linespec = "#L%d" % (lineno + 1)
    else:
        linespec = ""

    fn = os.path.relpath(fn, start=os.path.dirname(gwpy.__file__))
    if fn.startswith(os.path.pardir):
        return None

    tag = gwpy_version.git_tag or gwpy_version.git_branch
    return ("http://github.com/gwpy/gwpy/tree/%s/gwpy/%s%s"
            % (tag, fn, linespec))



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""This module redefines the matplotlib log scale to give a more
helpful set of major and minor ticks.
"""

from __future__ import division

from math import (ceil, floor, modf, log10)

import numpy

from matplotlib import rcParams
from matplotlib.axis import XAxis
from matplotlib.scale import (LogScale, register_scale)
from matplotlib.ticker import (is_decade, LogFormatterMathtext, LogLocator)

from .tex import float_to_latex
from .. import version

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


class GWpyLogFormatterMathtext(LogFormatterMathtext):
    """Format values for log axis.

    This `Formatter` extends the standard
    :class:`~matplotlib.ticker.LogFormatterMathtext` to print numbers
    in the range [0.01, 1000) normally, and all others via the
    `LogFormatterMathtext` output.
    """
    def __call__(self, x, pos=None):
        usetex = rcParams['text.usetex']
        if float(x).is_integer():
            x = int(x)
        if 0.01 <= x < 1000:
            if usetex:
                return '$%s$' % x
            else:
                return '$\mathdefault{%s}$' % x
        elif usetex and abs(log10(x)) > 2:
            return '$%s$' % float_to_latex(x, '%.2e')
        elif usetex:
            return '$%s$' % float_to_latex(x, '%.2g')
        else:
            return super(GWpyLogFormatterMathtext, self).__call__(x, pos=pos)


class MinorLogFormatterMathtext(GWpyLogFormatterMathtext):
    """Format minor tick labels on a log scale.

    This `Formatter` conditionally formats minor tick labels based on the
    number of major ticks visible when the formatter is called, either

    - no minor tick labels if two or more major ticks are visible
    - half-decade tick labels (0.5, 5, 50, ...) if only one major tick
      is visible
    - otherwise all minor ticks
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('labelOnlyBase', False)
        super(MinorLogFormatterMathtext, self).__init__(*args, **kwargs)

    def __call__(self, x, pos=None):
        """Format the minor tick label if less than 2 visible major ticks.
        """
        viewlim = self.axis.get_view_interval()
        loglim = numpy.log10(viewlim)
        majticks = numpy.arange(ceil(loglim[0]), floor(ceil(loglim[1])),
                                dtype=int)
        nticks = majticks.size
        halfdecade = nticks == 1 and modf(loglim[0])[0] < 0.7
        # if already two major ticks, don't need minor labels
        if nticks >= 2 or (halfdecade and not is_decade(x * 2, self._base)):
            return ''
        else:
            return super(MinorLogFormatterMathtext, self).__call__(x, pos=pos)


class CombinedLogFormatterMathtext(MinorLogFormatterMathtext):
    """Format major and minor ticks with a single `Formatter`.

    This is just a swap between the `MinorLogFormatterMathtext` and
    the GWpyLogFormatterMathtext` depending on whether the tick in
    question is a decade (0.1, 1, 10, 100, ...) or not.
    """
    def __call__(self, x, pos=None):
        if is_decade(x, self._base):
            return super(MinorLogFormatterMathtext, self).__call__(x, pos=pos)
        else:
            return super(CombinedLogFormatterMathtext, self).__call__(x,
                                                                      pos=pos)


class GWpyLogScale(LogScale):
    """GWpy version of the matplotlib `LogScale`.

    This scale overrides the default to use the new GWpy formatters
    for major and minor ticks.
    """
    def set_default_locators_and_formatters(self, axis):
        if isinstance(axis, XAxis):
            axis.set_tick_params(which='both', pad=7)
            axis.labelpad = 8
        axis.set_major_locator(LogLocator(self.base))
        axis.set_major_formatter(GWpyLogFormatterMathtext(self.base))
        axis.set_minor_locator(LogLocator(self.base, self.subs))
        axis.set_minor_formatter(MinorLogFormatterMathtext(self.base))

register_scale(GWpyLogScale)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for LIGO_LW tables.
"""

import re

import numpy

from . import lsctables
from .. import version

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version

EVENT_TABLES = (lsctables.SnglBurstTable,
                lsctables.MultiBurstTable,
                lsctables.SnglInspiralTable,
                lsctables.MultiInspiralTable,
                lsctables.SnglRingdownTable)

TIME_COLUMN = {
    lsctables.CoincInspiralTable.tableName: ('end_time', 'end_time_ns'),
    lsctables.CoincRingdownTable.tableName: ('start_time', 'start_time_ns'),
    lsctables.MultiBurstTable.tableName: ('peak_time', 'peak_time_ns'),
    lsctables.MultiInspiralTable.tableName: ('end_time', 'end_time_ns'),
    lsctables.SimBurstTable.tableName: ('time_geocent_gps',
                                        'time_geocent_gps_ns'),
    lsctables.SimInspiralTable.tableName: ('geocent_end_time',
                                           'geocent_end_time_ns'),
    lsctables.SimRingdownTable.tableName: ('geocent_start_time',
                                           'geocent_start_time_ns'),
    lsctables.SnglBurstTable.tableName: ('peak_time', 'peak_time_ns'),
    lsctables.SnglInspiralTable.tableName: ('end_time', 'end_time_ns'),
    lsctables.SnglRingdownTable.tableName: ('start_time', 'start_time_ns'),
}


def get_table_column(table, column, dtype=numpy.dtype(float)):
    """Extract a column from the given table.

    This method uses the following logic to determine how to extract a
    column:

    - if the table has a 'get_column' method, use that,
    - if column = 'time', use one of the ``get_start``, ``get_peak`` or
      ``get_end`` methods, depending on the table type, or
    - get the column using the name as given

    Parameters
    ----------
    table : :class:`glue.ligolw.table.Table`
        the LIGO_LW Table from which to extract a column.
    column : `str`
        the name of the column to find.
    dtype : :class:`~numpy.dtype`, `type`, optional
        the requested data type of the returned array.

    Returns
    -------
    array : :class:`numpy.ndarray`
        an array containing the data from the requested column
    """
    column = str(column).lower()
    if hasattr(table, 'get_%s' % column):
        return numpy.asarray(getattr(table, 'get_%s' % column)()).astype(dtype)
    elif column == 'time':
        if re.match('(sngl_inspiral|multi_inspiral)', table.tableName, re.I):
            return numpy.asarray(table.get_end()).astype(dtype)
        elif re.match('(sngl_burst|multi_burst)', table.tableName, re.I):
            return numpy.asarray(table.get_peak()).astype(dtype)
        elif re.match('(sngl_ring|multi_ring)', table.tableName, re.I):
            return numpy.asarray(table.get_start()).astype(dtype)
        elif re.match('sim_burst', table.tableName, re.I):
            return numpy.asarray(get_table_column(table, 'time_geocent_gps') +
                                 get_table_column(table, 'time_geocent_gps_ns')
                                 * 1e-9)
    if hasattr(table, 'get_column'):
        return numpy.asarray(table.get_column(column)).astype(dtype)
    else:
        return numpy.asarray(table.getColumnByName(column)).astype(dtype)


def get_row_value(row, attr):
    """Get the attribute value of a given LIGO_LW row.

    Parameters
    ----------
    row : `object`
        a row of a LIGO_LW `Table`.
    attr : `str`
        the name of the column attribute to retrieve.

    See Also
    --------
    get_table_column : for details on the column-name logic
    """
    attr = str(attr).lower()
    cname = row.__class__.__name__
    if hasattr(row, 'get_%s' % attr):
        return getattr(row, 'get_%s' % attr)()
    elif attr == 'time':
        if re.match('(Sngl|Multi)Inspiral', cname, re.I):
            return row.get_end()
        elif re.match('(Sngl|Multi)Burst', cname, re.I):
            return row.get_peak()
        elif re.match('(Sngl|Multi)Ring', cname, re.I):
            return row.get_start()
    else:
        return getattr(row, attr)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read events from an Omicron-format ROOT file.
"""

import sys

if sys.version_info[0] < 3:
    range = xrange

from astropy.io import registry

from glue.lal import (Cache, CacheEntry)

from .. import lsctables
from ... import version
from ...io.cache import open_cache
from ...time import LIGOTimeGPS
from ...utils import with_import

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version

OMICRON_COLUMNS = ['search', 'peak_time', 'peak_time_ns',
                   'start_time', 'start_time_ns',
                   'stop_time', 'stop_time_ns', 'duration',
                   'central_freq', 'peak_frequency',
                   'flow', 'fhigh', 'bandwidth',
                   'snr', 'amplitude', 'confidence']


def sngl_burst_from_root(tchain, columns=OMICRON_COLUMNS):
    """Build a `SnglBurst` from a ROOT `TChain`.

    Parameters
    ----------
    tchain : `ROOT.TChain`
        the ROOT object from which to read.
    columns : `list`
        a `list` of valid `LIGO_LW` column names to load.

    Returns
    -------
    burst : :class:`~glue.ligolw.lsctables.SnglBurst`
        a `SnglBurst` with attributes populated from the ROOT data.
    """
    t = lsctables.SnglBurst()
    t.search = u'omicron'

    # parse frequency data
    flow = tchain.fstart
    fhigh = tchain.fend
    if 'flow' in columns:
        t.flow = flow
    if 'fhigh' in columns:
        t.fhigh = fhigh
    if 'bandwidth' in columns:
        t.bandwidth = fhigh-flow
    if 'central_freq' in columns:
        t.central_freq = tchain.frequency
    if 'peak_frequency' in columns:
        t.peak_frequency = tchain.frequency

    # parse time data
    peak_time = LIGOTimeGPS(tchain.time)
    if 'time' in columns or 'peak_time' in columns:
        t.peak_time = peak_time.seconds
    if 'time' in columns or 'peak_time_ns' in columns:
        t.peak_time_ns = peak_time.nanoseconds
    start_time = LIGOTimeGPS(tchain.tstart)
    if 'start_time' in columns:
        t.start_time = start_time.seconds
    if 'start_time_ns' in columns:
        t.start_time_ns = start_time.nanoseconds
    stop_time = LIGOTimeGPS(tchain.tend)
    if 'stop_time' in columns:
        t.stop_time = stop_time.seconds
    if 'stop_time_ns' in columns:
        t.stop_time_ns = stop_time.nanoseconds
    if 'duration' in columns:
        t.duration = float(stop_time - start_time)

    # others
    if 'snr' in columns:
        t.snr = tchain.snr
    if 'amplitude' in columns:
        t.amplitude = tchain.snr ** 2 / 2.
    if 'confidence' in columns:
        t.confidence = tchain.snr

    return t


@with_import('ROOT')
def table_from_root(f, columns=OMICRON_COLUMNS, filt=None, nproc=1):
    """Build a `SnglBurstTable` from events in an Omicron ROOT file.

    Parameters
    ----------
    f : `file`, `str`, `CacheEntry`, `list`, `Cache`
        object representing one or more files. One of

        - an open `file`
        - a `str` pointing to a file path on disk
        - a formatted :class:`~glue.lal.CacheEntry` representing one file
        - a `list` of `str` file paths
        - a formatted :class:`~glue.lal.Cache` representing many files

    columns : `list`, optional
        list of column name strings to read, default all.
    filt : `function`, optional
        function by which to filt events. The callable must accept as
        input a `SnglBurst` event and return `True`/`False`.
    nproc : `int`, optional, default: 1
        number of parallel processes with which to distribute file I/O,
        default: serial process
    """
    # allow multiprocessing
    if nproc != 1:
        from .cache import read_cache
        return read_cache(f, lsctables.SnglBurstTable.tableName,
                          columns=columns, nproc=nproc, format='omicron')

    # format list of files
    if isinstance(f, CacheEntry):
        files = [f.path]
    elif isinstance(f, (str, unicode)) and f.endswith(('.cache', '.lcf')):
        files = open_cache(f).pfnlist()
    elif isinstance(f, (str, unicode)):
        files = f.split(',')
    elif isinstance(f, Cache):
        files = f.pfnlist()
    else:
        files = list(f)

    # read tree chain
    tree = ROOT.TChain('triggers')
    for filename in files:
        tree.Add(filename)

    # generate output
    out = lsctables.New(lsctables.SnglBurstTable, columns=columns)
    append = out.append

    # iterate over events
    nevents = tree.GetEntries()
    for i in range(nevents):
        tree.GetEntry()
        burst = sngl_burst_from_root(tree, columns=columns)
        if filt is None or filt(burst):
            append(burst)

    return out


def identify_omicron(*args, **kwargs):
    """Determine an input object as an Omicron-format ROOT file.
    """
    fp = args[3]
    if isinstance(fp, file):
        fp = fp.name
    elif isinstance(fp, CacheEntry):
        fp = fp.path
    # identify string
    if (isinstance(fp, (unicode, str)) and
            fp.endswith('root') and
            'omicron' in fp.lower()):
        return True
        # identify cache object
    else:
        return False


registry.register_reader('omicron', lsctables.SnglBurstTable, table_from_root)
registry.register_identifier('omicron', lsctables.SnglBurstTable,
                             identify_omicron)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Helper functions for plotting data with matplotlib and LAL
"""

import numpy
import itertools
import re

from . import rcParams
from .. import version

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version

rUNDERSCORE = re.compile(r'(?<!\\)_')

# groups of input parameters (for passing to Plot() and subclasses)
FIGURE_PARAMS = [
    'figsize',
]
AXES_PARAMS = [
    'projection', 'sharex', 'sharey', 'xlim', 'ylim', 'xlabel', 'ylabel',
    'xscale', 'yscale', 'title',
]
LINE_PARAMS = [
    'linewidth', 'linestyle', 'color', 'label', 'alpha',
]
COLLECTION_PARAMS = [
    'cmap', 'vmin', 'vmax', 'marker', 's', 'norm',
]
ARTIST_PARAMS = set(LINE_PARAMS + COLLECTION_PARAMS)
LEGEND_PARAMS = [
    'loc', 'borderaxespad', 'ncol',
]


def float_to_latex(x, format="%.2g"):
    """Convert a floating point number into a TeX representation.

    In particular, scientific notation is handled gracefully: e -> 10^

    @code
    >>> float_to_latex(10)
    '$10$'
    >>> float_to_latex(1000)
    r'$10^{3}$'
    >>> float_to_latex(123456789)
    r'$1.2\times 10^{8}$'

    @returns a TeX format string (with math-mode dollars ($))
    """
    base_str = format % x
    if "e" not in base_str:
        return "$%s$" % base_str
    mantissa, exponent = base_str.split("e")
    exponent = exponent.lstrip("0+")
    if mantissa == "1":
        return r"$10^{%s}$" % exponent
    else:
        return r"$%s\times 10^{%s}$" % (mantissa, exponent)


def color_cycle(colors=None):
    """An infinite iterator of the given (or default) color cycle.
    """
    if colors:
        return itertools.cycle(colors)
    else:
        return itertools.cycle(rcParams['axes.color_cycle'])


def marker_cycle(markers=None):
    if markers:
        return itertools.cycle(markers)
    else:
        return itertools.cycle(('o', 'x', '+', '^', 'D', 'H', '1'))



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2013-2015)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>

"""Calculating and plotting a `SpectralVariance` histogram
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__currentmodule__ = 'gwpy.spectrum'

# In order to generate a `SpectralVariance` histogram we need to import the
# `~gwpy.timeseries.TimeSeries` and :meth:`~gwpy.timeseries.TimeSeries.get`
# the data:
from gwpy.timeseries import TimeSeries
llo = TimeSeries.get(
    'L1:LDAS-STRAIN,rds', 'August 1 2010', 'August 1 2010 00:10')

# We can then call the :meth:`~gwpy.timeseries.TimeSeries.spectral_variance`
# method of the ``llo`` `~gwpy.timeseries.TimeSeries`:
variance = llo.spectral_variance(1, log=True, low=1e-24, high=1e-19, nbins=100)

# We can then :meth:`~SpectralVariance.plot` the `SpectralVariance`
plot = variance.plot(norm='log', vmin=0.5, vmax=100)
ax = plot.gca()
ax.grid()
ax.set_xlim(40, 4096)
ax.set_ylim(1e-24, 1e-19)
ax.set_xlabel('Frequency [Hz]')
ax.set_ylabel(r'GW ASD [strain/\rtHz]')
ax.set_title('LIGO Livingston Observatory sensitivity variance')
plot.show()



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Plotting a whitened `Spectrogram`

I would like to study the gravitational wave strain spectrogram around the time of an interesting simulated signal during the last science run (S6).
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__currentmodule__ = 'gwpy.spectrogram'

# As with :doc:`previous example <plot>`, we import the
# `~gwpy.timeseries.TimeSeries` class,
# :meth:`~gwpy.timeseries.TimeSeries.get` the data, and calculate a 
# `Spectrogram`
from gwpy.timeseries import TimeSeries
gwdata = TimeSeries.get('H1:LDAS-STRAIN,rds', 'September 16 2010 06:40', 'September 16 2010 06:50')
specgram = gwdata.spectrogram(5, fftlength=2, overlap=1) ** (1/2.)

# To whiten the `specgram` we can use the :meth:`~Spectrogram.ratio` method
# to divide by the overall median:
medratio = specgram.ratio('median')

# Finally, we make a plot:
plot = medratio.plot(norm='log', vmin=0.1, vmax=10)
plot.set_yscale('log')
plot.set_ylim(40, 4096)
plot.add_colorbar(label='Amplitude relative to median')
plot.show()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Input/Output utilities for LAL Cache files.
"""

from __future__ import division
from math import ceil
from multiprocessing import (cpu_count, Process, Queue as ProcessQueue)
from six import string_types
import warnings

from glue.lal import (Cache, CacheEntry)
from glue.ligolw.table import Table

from astropy.io.registry import _get_valid_format

from .. import version

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


def open_cache(lcf):
    """Read a LAL-format cache file into memory as a
    :class:`glue.lal.Cache`.
    """
    if isinstance(lcf, file):
        return Cache.fromfile(lcf)
    else:
        with open(lcf, 'r') as f:
            return Cache.fromfile(f)


def identify_cache_file(*args, **kwargs):
    """Determine an input object as either a LAL-format cache file.
    """
    cachefile = args[3]
    if isinstance(cachefile, file):
        cachefile = cachefile.name
    # identify string
    if (isinstance(cachefile, string_types) and
            cachefile.endswith(('.lcf', '.cache'))):
        return True
    # identify cache object
    else:
        return False


def identify_cache(*args, **kwargs):
    """Determine an input object as a :class:`glue.lal.Cache`
    """
    cacheobj = args[3]
    if isinstance(cacheobj, Cache):
        return True
    else:
        return False


def file_list(flist):
    """Parse a number of possible input types into a list of filepaths.
    """
    # format list of files
    if isinstance(flist, file):
        return [flist.name]
    elif isinstance(flist, CacheEntry):
        return [flist.path]
    elif (isinstance(flist, string_types) and
          flist.endswith(('.cache', '.lcf'))):
        return open_cache(flist).pfnlist()
    elif isinstance(flist, string_types):
        return flist.split(',')
    elif isinstance(flist, Cache):
        return flist.pfnlist()
    elif isinstance(flist, (list, tuple)):
        return flist
    else:
        return list([flist])


# ----------------------------------------------------------------------------
# generic multiprocessing from caches

def read_cache(cache, target, nproc, post, *args, **kwargs):
    """Read arbitrary data from a cache file

    Parameters
    ----------
    cache : :class:`glue.lal.Cache`, `str`
        cache of files files, or path to a LAL-format cache file
        on disk.
    target : `type`
        target class to read into.
    nproc : `int`
        number of individual processes to use.
    post : `function`
        function to post-process output object before returning.
        The output of this method will be returns, so in-place operations
        must return the object.
    *args
        other positional arguments to pass to the target.read()
        classmethod.
    **kwargs
        keyword arguments to pass to the target.read() classmethod.

    Returns
    -------
    data : target
        an instance of the target class, seeded with data read from
        the cache.

    Notes
    -----
    The returned object is constructed from the output of each
    sub-process via the '+=' in-place addition operator.

    If the input cache is indeed a :class:`~glue.lal.Cache` object,
    the sub-processes will be combined in time order, otherwise the ordering
    is given by the order of entries in the input cache (for example,
    if it is a simple `list` of files).

    .. warning::

       no protection is given against overloading the host, for example,
       no checks are done to ensure that ``nproc`` is less than the number
       of available cores.

       High values of ``nproc`` should be used at the users discretion,
       the GWpy team accepts to liability for loss as a result of abuse
       of this feature.
    """
    # read the cache
    if isinstance(cache, (file, unicode, str)):
        cache = open_cache(cache)
    if isinstance(cache, Cache):
        cache.sort(key=lambda ce: ce.segment[0])

    # force one file per process minimum
    nproc = min(nproc, len(cache))
    if nproc > cpu_count():
        warnings.warn("Using %d processes on a %d-core machine is "
                      "unrecommended...but not forbidden."
                      % (nproc, cpu_count()))

    # work out underlying data type
    try:
        kwargs.setdefault(
            'format', _get_valid_format('read', target, None,
                                        None, (cache[0],), {}))
    # if empty, put anything, since it doesn't matter
    except IndexError:
        kwargs.setdefault('format', 'ligolw')
    except Exception:
        if 'format' not in kwargs:
            raise

    if nproc <= 1:
        return target.read(cache, *args, **kwargs)

    # define how to read each sub-cache
    def _read(q, sc, i):
        try:
            q.put((i, target.read(sc, *args, **kwargs)))
        except Exception as e:
            q.put(e)

    # separate cache into parts
    fperproc = int(ceil(len(cache) / nproc))
    subcaches = [cache.__class__(cache[i:i+fperproc]) for
                 i in range(0, len(cache), fperproc)]

    # start all processes
    queue = ProcessQueue(nproc)
    proclist = []
    for i, subcache in enumerate(subcaches):
        process = Process(target=_read, args=(queue, subcache, i))
        process.daemon = True
        proclist.append(process)
        process.start()

    # get data and block
    pout = []
    for i in range(len(proclist)):
        result = queue.get()
        if isinstance(result, Exception):
            raise result
        pout.append(result)
    for process in proclist:
        process.join()

    # combine and return
    data = zip(*sorted(pout, key=lambda out: out[0]))[1]
    try:
        if issubclass(target, Table):
            out = data[0]
        else:
            out = data[0].copy()
    except AttributeError:
        out = data[0]
    for datum in data[1:]:
        out += datum

    if post:
        return post(out)
    else:
        return out


def read_cache_factory(target):
    """Generate a read_cache method specific to a given target class.

    Parameters
    ----------
    target : `type`
        `class` object to bind to returned method

    Returns
    -------
    read_target_cache : `function`
        newly-minted method taking arbitrary arguments and keyword
        arguments, and handling type associations and multiprocessing.
    """
    def _read(f, *args, **kwargs):
        nproc = kwargs.pop('nproc', 1)
        post = kwargs.pop('post', None)
        return read_cache(f, target, nproc, post, *args, **kwargs)
    return _read


def cache_segments(*caches, **kwargs):
    """Build a `SegmentList` of data availability for these `Caches`

    Parameters
    ----------
    *cache : `~glue.lal.Cache`
        one of more frame file caches describing files on disk
    on_missing : `str`
        what to do if files in a `Cache` are not found on disk, one of

        - "warn": print a warning message saying how many files
                  are missing out of the total checked [DEFAULT].
        - "error": raise an exception if any are missing
        - "ignore": do nothing

    Returns
    -------
    segments : `~gwpy.segments.SegmentList`
        a list of segments for when data should be available
    """
    from ..segments import SegmentList
    on_missing = kwargs.pop('on_missing', 'warn')
    if kwargs:
        raise TypeError("%r is an invalid keyword for cache_segments"
                        % kwargs.keys()[0])
    out = SegmentList()
    for cache in caches:
        found, _ = cache.checkfilesexist(on_missing=on_missing)
        out.extend(e.segment for e in found)
    return out.coalesce()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read/write segment XML in LIGO_LW format into DataQualityFlags
"""

import datetime
from six import string_types

from glue.lal import LIGOTimeGPS
from glue.ligolw.ligolw import (Document, LIGO_LW)
from glue.ligolw.utils import (write_filename, write_fileobj)
from glue.ligolw.utils.ligolw_add import ligolw_add


from astropy.time import Time
from astropy.io import registry

from ... import version
from ...io.utils import GzipFile
from ...io.ligolw import (identify_ligolw, GWpyContentHandler)
from ...io.cache import file_list
from ...segments import (Segment, DataQualityFlag, DataQualityDict)
from ...table import lsctables

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version


def read_flag_dict(f, flags=None, gpstype=LIGOTimeGPS, coalesce=False,
                   contenthandler=GWpyContentHandler, nproc=1):
    """Read segments for the given flag from the LIGO_LW XML file.

    Parameters
    ----------
    fp : `str`
        path of XML file to read.
    flags : `list`, `None`, optional
        list of flags to read or `None` to read all into a single
        `DataQualityFlag`.

    Returns
    -------
    flagdict : :class:`~gwpy.segments.flag.DataQualityDict`
        a new `DataQualityDict` of `DataQualityFlag` entries with ``active``
        and ``known`` segments seeded from the XML tables in the given
        file ``fp``.
    """
    if nproc != 1:
        return DataQualityDict.read(f, flags, coalesce=coalesce,
                                    gpstype=gpstype,
                                    contenthandler=contenthandler,
                                    format='cache', nproc=nproc)

    # generate Document and populate
    xmldoc = Document()
    files = [fp.name if isinstance(fp, (file, GzipFile))
             else fp for fp in file_list(f)]
    ligolw_add(xmldoc, files, non_lsc_tables_ok=True,
               contenthandler=contenthandler)

    # read segment definers and generate DataQualityFlag object
    seg_def_table = lsctables.SegmentDefTable.get_table(xmldoc)

    # find flags
    if isinstance(flags, (unicode, str)):
        flags = flags.split(',')
    out = DataQualityDict()
    id_ = dict()
    if flags is not None and len(flags) == 1 and flags[0] is None:
        out[None] = DataQualityFlag()
        id_[None] = []
    for row in seg_def_table:
        ifos = row.get_ifos()
        name = row.name
        if ifos and name:
            name = ':'.join([''.join(row.get_ifos()), row.name])
            if row.version is not None:
                name += ':%d' % row.version
        else:
            name = None
        if flags is None or name in flags:
            out[name] = DataQualityFlag(name)
            try:
                id_[name].append(row.segment_def_id)
            except (AttributeError, KeyError):
                id_[name] = [row.segment_def_id]
    if flags is None and not len(out.keys()):
        raise RuntimeError("No segment definitions found in file.")
    elif flags is not None and len(out.keys()) != len(flags):
        for flag in flags:
            if flag not in out:
                raise ValueError("No segment definition found for flag=%r "
                                 "in file." % flag)
    # read segment summary table as 'known'
    seg_sum_table = lsctables.SegmentSumTable.get_table(xmldoc)
    for row in seg_sum_table:
        for flag in out:
            if not id_[flag] or row.segment_def_id in id_[flag]:
                try:
                    s = row.get()
                except AttributeError:
                    s = row.start_time, row.end_time
                out[flag].known.append(Segment(gpstype(s[0]), gpstype(s[1])))
    for dqf in out:
        if coalesce:
            out[dqf].coalesce()
    # read segment table as 'active'
    seg_table = lsctables.SegmentTable.get_table(xmldoc)
    for row in seg_table:
        for flag in out:
            if not id_[flag] or row.segment_def_id in id_[flag]:
                try:
                    s = row.get()
                except AttributeError:
                    s = row.start_time, row.end_time
                out[flag].active.append(Segment(gpstype(s[0]), gpstype(s[1])))
    for dqf in out:
        if coalesce:
            out[dqf].coalesce()
    return out


def read_flag(fp, flag=None, **kwargs):
    """Read a single `DataQualityFlag` from a LIGO_LW XML file
    """
    return read_flag_dict(fp, flags=flag, **kwargs).values()[0]


def write_ligolw(flag, fobj, **kwargs):
    """Write this `DataQualityFlag` to XML in LIGO_LW format
    """
    # if given a Document, just add data
    if isinstance(fobj, Document):
        return write_to_xmldoc(flag, fobj, **kwargs)
    # otherwise build a new Document
    xmldoc = Document()
    xmldoc.appendChild(LIGO_LW())
    # TODO: add process information
    write_to_xmldoc(flag, xmldoc)
    # and write
    if isinstance(fobj, string_types):
        return write_filename(xmldoc, fobj, gz=fobj.endswith('.gz'))
    else:
        return write_fileobj(xmldoc, fobj, gz=fobj.name.endswith('.gz'))


def write_to_xmldoc(flags, xmldoc, process_id=None):
    """Write this `DataQualityFlag` to the given LIGO_LW Document
    """
    if isinstance(flags, DataQualityFlag):
        flags = {flags.name: flags}

    # write SegmentDefTable
    try:
        segdeftab = lsctables.SegmentDefTable.get_table(xmldoc)
    except ValueError:
        segdeftab = lsctables.New(lsctables.SegmentDefTable,
                                  columns=['ifos', 'name', 'version',
                                           'comment', 'insertion_time',
                                           'segment_def_id', 'process_id'])
        xmldoc.childNodes[-1].appendChild(segdeftab)
    # make SegmentSumTable
    try:
        segsumtab = lsctables.SegmentSumTable.get_table(xmldoc)
    except ValueError:
        segsumtab = lsctables.New(lsctables.SegmentSumTable,
                                  columns=['segment_def_id', 'start_time',
                                           'start_time_ns', 'end_time',
                                           'end_time_ns', 'comment',
                                           'segment_sum_id', 'process_id'])
        xmldoc.childNodes[-1].appendChild(segsumtab)
    # write SegmentTable
    try:
        segtab = lsctables.SegmentTable.get_table(xmldoc)
    except ValueError:
        segtab = lsctables.New(lsctables.SegmentTable,
                               columns=['process_id', 'segment_id',
                                        'segment_def_id', 'start_time',
                                        'start_time_ns', 'end_time',
                                        'end_time_ns'])
        xmldoc.childNodes[-1].appendChild(segtab)

    # write flags to tables
    for flag in flags.itervalues():
        # segment definer
        segdef = lsctables.SegmentDef()
        segdef.set_ifos([flag.ifo])
        segdef.name = flag.tag
        segdef.version = flag.version
        segdef.comment = flag.description
        segdef.insertion_time = int(Time(datetime.datetime.now(),
                                         scale='utc').gps)
        segdef.segment_def_id = lsctables.SegmentDefTable.get_next_id()
        segdef.process_id = process_id
        segdeftab.append(segdef)

        # write segment summary (known segments)
        for vseg in flag.known:
            segsum = lsctables.SegmentSum()
            segsum.segment_def_id = segdef.segment_def_id
            segsum.set(map(LIGOTimeGPS, map(float, vseg)))
            segsum.comment = None
            segsum.segment_sum_id = lsctables.SegmentSumTable.get_next_id()
            segsum.process_id = process_id
            segsumtab.append(segsum)

        # write segment table (active segments)
        for aseg in flag.active:
            seg = lsctables.Segment()
            seg.segment_def_id = segdef.segment_def_id
            seg.set(map(LIGOTimeGPS, map(float, aseg)))
            seg.segment_id = lsctables.SegmentTable.get_next_id()
            seg.process_id = process_id
            segtab.append(seg)

    return xmldoc


# register methods for DataQualityDict
registry.register_reader('ligolw', DataQualityFlag, read_flag)
registry.register_writer('ligolw', DataQualityFlag, write_ligolw)
registry.register_identifier('ligolw', DataQualityFlag, identify_ligolw)

# register methods for DataQualityDict
registry.register_reader('ligolw', DataQualityDict, read_flag_dict)
registry.register_writer('ligolw', DataQualityDict, write_ligolw)
registry.register_identifier('ligolw', DataQualityDict, identify_ligolw)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) Joseph Areeda (2015)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.
#

"""Base class for cli plot products, common code for arguments, data transfer,
plot annotation:

Abstract methods (must be overridden by actual products):
    get_action()     - return the string used as "action" in command line
    init_cli(parser) - set up the argument list for this product

Implemented methods (used by multiple products):
    arg_chan(parser) - add definition of a list TimeSeries objects
    arg_freq(parser) - add parameters for fft based plots
    arg_time(parser) - add parameters for time based plots
    arg_plot(parser) - add plot annotations (labels, legends..)
"""

import abc
import sys
import re

from argparse import ArgumentError

from gwpy.timeseries import TimeSeries


class CliProduct(object):
    """Base class for all cli plot products"""

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self.min_timeseries = 1     # datasets needed for this product
        self.xaxis_type = 'uk'      # scaling hints, set by individual actions
        self.xaxis_is_freq = False  # x is going to be frequency or time
        self.yaxis_type = 'uk'      # x and y axis types must be set
        self.iaxis_type = None      # intensity axis (colors) may be missing
        self.chan_list = []         # list of channel names
        self.start_list = []        # list of start times as GPS seconds
        self.dur = 0                # one duration in secs for all start times
        self.timeseries = []        # time series objects after transfer
        self.time_groups = []       # lists of indexes into time series,
        self.minMax = []
        self.verbose = 1
        self.secpfft = 1
        self.overlap = 0.5
        # describe the actual plotted limits
        self.fmin = 0
        self.fmax = 1
        self.ymin = 0
        self.ymax = 0
        self.xmin = 0
        self.xmax = 0
        self.imin = 0
        self.imax = 0
        # image geometry
        self.width = 0
        self.height = 0
        self.xinch = 12
        self.yinch = 7.68
        self.dpi = 100

        self.is_freq_plot = False
        self.n_datasets = 0
        self.filter = ''        # string for annotation if we filtered data
        self.plot = 0           # plot object
        self.ax = 0             # current axis object from plot

# ------Abstract metods------------
    @abc.abstractmethod
    def get_action(self):
        """Return the string used as "action" on command line."""
        return

    @abc.abstractmethod
    def init_cli(self, parser):
        """Set up the argument list for this product"""
        return

    @abc.abstractmethod
    def gen_plot(self, args):
        """Generate the plot from time series and arguments"""
        return

    @abc.abstractmethod
    def get_ylabel(self, args):
        """Text for y-axis label"""
        return

    @abc.abstractmethod
    def get_title(self):
        """Start of default super title, first channel is appended to it"""
        return

# --------Defaults for overridable methods if product differs from default
    def get_min_datasets(self):
        """Override if plot requires more than 1 dataset.
        eg: coherence requires 2"""
        return 1

    def get_max_datasets(self):
        """Override if plot has a maximum number of datasets.
        eg: spectrogram only handles 1"""
        return 16  # arbitrary max

    def is_image(self):
        """Override if plot is image type, eg: spectrogram"""
        return False

    def freq_is_y(self):
        """Override if frequency is on y-axis like spectrogram"""
        return False

    def get_xlabel(self):
        """Override if you have a better label.
        default is to usw gwpy's default label"""
        return ''

    def get_color_label(self):
        """All products with a color bar should override this"""
        return 'action does not label color bar (it should)'

    def get_sup_title(self):
        """Override if default lacks critical info"""
        return self.get_title() + self.timeseries[0].channel.name

# ------Helper functions
    def log(self, level, msg):
        """print log message if verbosity is set high enough"""
        if self.verbose >= level:
            print msg
        return


# ------Argparse methods

    def arg_chan(self, parser):
        """Allow user to specify list of channel names,
        list of start gps times and single duration"""
        parser.add_argument('--start', nargs='+', action='append',
                            required=True, help='Starting GPS times(required)')
        parser.add_argument('--duration', default=10,
                            help='Duration (seconds) [10]')
        parser.add_argument('-c', '--framecache',
                            help='use .gwf files in cache not NDS2,' +
                                 ' default use NDS2')
        parser.add_argument('--highpass',
                            help='frequency for high pass butterworth,' +
                                 ' default no filter')
        return

    def arg_chan1(self, parser):
        # list of channel names when only 1 is required
        parser.add_argument('--chan', nargs='+', action='append',
                            required=True, help='One or more channel names.')
        self.arg_chan(parser)
        return

    def arg_chan2(self, parser):
        """list of channel names when at least 2 are required"""
        parser.add_argument('--chan', nargs='+', action='append',
                            required=True,
                            help='Two or more channels or times, first' +
                                 ' one is compared to all the others')
        parser.add_argument('--ref',
                            help='Reference channel against which ' +
                                 'others will be compared')

        self.arg_chan(parser)

        return

    def arg_freq(self, parser):
        """Parameters for FFT based plots, with Spectral defaults"""
        self.is_freq_plot = True
        parser.add_argument('--secpfft', default='1.0',
                            help='length of fft in seconds ' +
                                 'for each calculation, default = 1.0')
        parser.add_argument('--overlap', default='0.5',
                            help='Overlap as fraction [0-1), default=0.5')
        return

    def arg_freq2(self, parser):
        """Parameters for FFT based plots, with Coherencegram defaults"""
        self.is_freq_plot = True
        parser.add_argument('--secpfft', default='0.5',
                            help='length of fft in seconds ' +
                                 'for each calculation, default=0.5')
        parser.add_argument('--overlap', default='0.9',
                            help='Overlap as fraction [0-1), default=0.9')
        return

    def arg_plot(self, parser):
        """Add arguments common to all plots"""
        parser.add_argument('-g', '--geometry', default='1200x600',
                            help='size of resulting image WxH, ' +
                                 'default: %(default)s')
        parser.add_argument('--interactive', action='store_true',
                            help='when running from ipython ' +
                                 'allows experimentation')
        parser.add_argument('--title', action='append',
                            help='One or more title lines')
        parser.add_argument('--suptitle',
                            help='1st title line (larger than the others)')
        parser.add_argument('--xlabel', help='x axis text')
        parser.add_argument('--ylabel', help='y axis text')
        parser.add_argument('--out',
                            help='output filename, type=ext (png, pdf, ' +
                                 'jpg), default=gwpy.png')
        # legends match input files in position are displayed if specified.
        parser.add_argument('--legend', nargs='*', action='append',
                            help='strings to match data files')
        parser.add_argument('--nolegend', action='store_true',
                            help='do not display legend')
        parser.add_argument('--nogrid', action='store_true',
                            help='do not display grid lines')

        return

    def arg_ax_x(self, parser):
        """X-axis is called X. Do not call this
        one call arg_ax_linx or arg_ax_logx"""
        parser.add_argument('--xmin', help='min value for X-axis')
        parser.add_argument('--xmax', help='max value for X-axis')
        return

    def arg_ax_linx(self, parser):
        """X-axis is called X and defaults to linear"""
        self.xaxis_type = 'linx'
        parser.add_argument('--logx', action='store_true',
                            help='make X-axis logarithmic, default=linear')
        parser.add_argument('--epoch',
                            help='center X axis on this GPS time. ' +
                                 'Incompatible with logx')
        self.arg_ax_x(parser)
        return

    def arg_ax_logx(self, parser):
        """X-axis is called X and defaults to logarithmic"""
        self.xaxis_type = 'logx'
        parser.add_argument('--nologx', action='store_true',
                            help='make X-axis linear, default=logarithmic')
        self.arg_ax_x(parser)
        return

    def arg_ax_lf(self, parser):
        """One of this  axis is frequency and logarthmic"""
        parser.add_argument('--nologf', action='store_true',
                            help='make frequency axis linear, ' +
                                 'default=logarithmic')
        parser.add_argument('--fmin', help='min value for frequency axis')
        parser.add_argument('--fmax', help='max value for frequency axis')
        return

    def arg_ax_int(self, parser):
        """Images have an intensity axis"""
        parser.add_argument('--imin',
                            help='min pixel value in resulting image')
        parser.add_argument('--imax',
                            help='max pixek value in resulting image')
        return

    def arg_ax_intlin(self, parser):
        """Intensity (colors) default to linear"""
        self.iaxis = 'lini'
        parser.add_argument('--logcolors', action='store_true',
                            help='set intensity scale of image ' +
                                 'to logarithmic, default=linear')
        self.arg_ax_int(parser)
        return

    def arg_ax_intlog(self, parser):
        """Intensity (colors) default to log"""
        self.iaxis = "logi"
        parser.add_argument('--lincolors', action='store_true',
                            help='set intensity scale of image to linear, '
                                 'default=logarithmic')
        self.arg_ax_int(parser)
        return

    def arg_ax_xlf(self, parser):
        """X-axis is called F and defaults to log"""
        self.xaxis_type = 'logf'
        self.arg_ax_lf(parser)
        return

    def arg_ax_ylf(self, parser):
        """Y-axis is called Frequency and defaults to log"""
        self.yaxis_type = 'logf'
        self.arg_ax_lf(parser)
        return

    def arg_ax_y(self, parser):
        """Y-axis limits.  Do not call this one
        use arg_ax_liny or arg_ax_logy"""
        parser.add_argument('--ymin', help='fix min value for yaxis' +
                                           ' defaults to min of data')
        parser.add_argument('--ymax', help='max value for y-axis ' +
                                           'default to max of data')
        return

    def arg_ax_liny(self, parser):
        """Y-axis is called Y and defaults to linear"""
        self.yaxis_type = 'liny'
        parser.add_argument('--logy', action='store_true',
                            help='make Y-axis logarithmic, default=linear')
        self.arg_ax_y(parser)
        return

    def arg_ax_logy(self, parser):
        """Y-axis is called Y and defaults to log"""
        self.yaxis_type = 'logy'
        parser.add_argument('--nology', action='store_true',
                            help='make Y-axis linear, default=logarthmic')
        self.arg_ax_y(parser)
        return

    def arg_imag(self, parser):
        """Add arguments for image based plots like spectrograms"""
        parser.add_argument('--nopct', action='store_true',
                            help='up and lo are pixel values, ' +
                                 'default=percentile if not normalized')
        parser.add_argument('--nocolorbar', action='store_true',
                            help='hide the color bar')
        parser.add_argument('--norm', action='store_true',
                            help='Display the ratio of each fequency ' +
                                 'bin to the mean of that frequency')
        return

# -------Data transfer methods

    def getTimeSeries(self, arg_list):
        """Verify and interpret arguments to get all
        TimeSeries objects defined"""

        # retrieve channel data from NDS as a TimeSeries
        for chans in arg_list.chan:
            for chan in chans:
                if chan not in self.chan_list:
                    self.chan_list.append(chan)

        if len(self.chan_list) < self.min_timeseries:
            raise ArgumentError(
                'A minimum of %d channels must be specified for this product'
                % self.min_timeseries)

        if len(arg_list.start) > 0:
            for start_arg in arg_list.start:
                if type(start_arg) is list:
                    for starts in start_arg:
                        if isinstance(starts, basestring):
                            starti = int(starts)

                        elif starts is list:
                            for start_str in starts:
                                starti = int(start_str)
                        # ignore duplicates (to make it easy for ldvw)
                        if starti not in self.start_list:
                            self.start_list.append(starti)
                else:
                    self.start_list.append(int(start_arg))
        else:
            raise ArgumentError('No start times specified')

        # Verify the number of datasets specified is valid for this plot
        self.n_datasets = len(self.chan_list) * len(self.start_list)
        if self.n_datasets < self.get_min_datasets():
            raise ArgumentError(
                '%d datasets are required for this plot but only %d are '
                'supplied' % (self.get_min_datasets(), self.n_datasets))

        if self.n_datasets > self.get_max_datasets():
            raise ArgumentError(
                'A maximum of %d datasets allowed for this plot but %d '
                'specified' % (self.get_max_datasets(), self.n_datasets))

        if arg_list.duration:
            self.dur = int(arg_list.duration)
        else:
            self.dur = 10

        verb = self.verbose > 1

        # determine how we're supposed get our data
        source = 'NDS2'
        frame_cache = False

        if arg_list.framecache:
            source = 'frames'
            frame_cache = arg_list.framecache

        # set up filter parameters for all channels
        highpass = 0
        if arg_list.highpass:
            highpass = float(arg_list.highpass)
            self.filter += "highpass(%.1f) " % highpass

        # Get the data from NDS or Frames
        # time_groups is a list of timeseries index grouped by
        # start time for coherence like plots
        self.time_groups = []
        for start in self.start_list:
            time_group = []
            for chan in self.chan_list:
                if verb:
                    print 'Fetching %s %d, %d using %s' % \
                          (chan, start, self.dur, source)
                if frame_cache:
                    data = TimeSeries.read(frame_cache, chan, start=start,
                                           end=start+self.dur)
                else:
                    data = TimeSeries.fetch(chan, start, start+self.dur,
                                            verbose=verb)

                if highpass > 0:
                    data = data.highpass(highpass)

                self.timeseries.append(data)
                time_group.append(len(self.timeseries)-1)
            self.time_groups.append(time_group)

        # report what we have if they asked for it
        self.log(3, ('Channels: %s' % self.chan_list))
        self.log(3, ('Start times: %s, duration' % self.start_list, self.dur))
        self.log(3, ('Number of time series: %d' % len(self.timeseries)))

        if len(self.timeseries) != self.n_datasets:
            self.log(0, ('%d datasets requested but only %d transfered' %
                         (self.n_datasets, len(self.timeseries))))
            if len(self.timeseries) > self.get_min_datasets():
                self.log(0, 'Proceeding with the data that was transferred.')
            else:
                self.log(0, 'Not enough data for requested plot.')
                sys.exit(2)
        return

# ---- Plotting methods

    def show_plot_info(self):
        self.log(3, ('X-scale: %s, Y-scale: %s' % (self.plot.get_xscale(),
                                                   self.plot.get_yscale())))
        self.log(3, ('X-limits %s, Y-limits %s' % (self.plot.xlim,
                                                   self.plot.ylim)))

    def config_plot(self, arg_list):
        """Configure global plot parameters"""
        from matplotlib import rcParams

        # set rcParams
        rcParams.update({
            'figure.dpi': 100.,
            'font.family': 'sans-serif',
            'font.size': 16.,
            'font.weight': 'book',
            'legend.loc': 'best',
            'lines.linewidth': 1.5,
            'text.usetex': 'true',
            'agg.path.chunksize': 10000,
        })

        # determine image dimensions (geometry)
        self.width = 1200
        self.height = 768
        if arg_list.geometry:
            try:
                self.width, self.height = map(float,
                                              arg_list.geometry.split('x', 1))
                self.height = max(self.height, 500)
            except (TypeError, ValueError) as e:
                e.args = ('Cannot parse --geometry as WxH, e.g. 1200x600',)
                raise

        self.dpi = rcParams['figure.dpi']
        self.xinch = self.width / self.dpi
        self.yinch = self.height / self.dpi
        rcParams['figure.figsize'] = (self.xinch, self.yinch)
        return

    def setup_xaxis(self, arg_list):
        """Handle scale and limits of X-axis by type"""

        xmin = 0        # these will be set by x min, max or f min, max
        xmax = 1
        scale = 'linear'
        if self.xaxis_type == 'linx' or self.xaxis_type == 'logx':
            # handle time on X-axis
            xmin = self.xmin
            xmax = self.xmax
            if arg_list.xmin:
                al_xmin = float(arg_list.xmin)
                if self.xaxis_is_freq:
                    xmin = al_xmin      # frequency specified
                elif al_xmin <= 1e8:
                    xmin = self.xmin + al_xmin     # time specified as
                    # seconds relative to start GPS
                else:
                    xmin = al_xmin      # time specified as absolute GPS
            if arg_list.xmax:
                al_xmax = float(arg_list.xmax)
                if self.xaxis_is_freq:
                    xmax = al_xmax
                elif al_xmax <= 9e8:
                    xmax = self.xmin + al_xmax
                else:
                    xmax = al_xmax

            if self.xaxis_type == 'linx':
                epoch = None
                if arg_list.epoch:
                    epoch = float(arg_list.epoch)
                    self.log(3, ('Epoch set to %.2f' % epoch))

                scale = False
                self.ax.set_xscale('auto-gps')

                if epoch is not None:
                    # note zero is also false
                    if epoch > 0 and epoch < 1e8:
                        epoch += self.xmin       # specified as seconds
                        self.ax.set_epoch(epoch)
                    elif epoch == 0:
                        self.ax.set_xscale('gps')
                        self.ax.set_epoch(0)
                if arg_list.logx:
                    scale = 'log'
                elif not (self.get_xlabel() or arg_list.xlabel):
                    # duplicate default label except use parens not brackets
                    xscale = self.ax.xaxis._scale
                    if self.ax.get_xscale() == 'auto-gps':
                        epoch = xscale.get_epoch()
                        if epoch is None:
                            arg_list.xlabel = 'GPS Time'
                if self.ax.get_xscale() == 'gps':
                    for l in self.ax.xaxis.get_ticklabels():
                        l.set_rotation(25)
                        l.set_ha('right')
            elif self.xaxis_type == 'logx':
                if arg_list.nologx:
                    scale = 'linear'
                else:
                    scale = 'log'
        elif self.xaxis_type == 'logf':
            # Handle frequency on the X-axis
            xmin = self.fmin
            xmax = self.fmax

            scale = 'log'
            if arg_list.nologf:
                scale = 'linear'
            if arg_list.fmin:
                xmin = float(arg_list.fmin)
            if arg_list.fmax:
                xmax = float(arg_list.fmax)
        else:
            raise AssertionError('X-axis type [%s] is unknown' %
                                 self.xaxis_type)

        if scale:
            self.ax.set_xscale(scale)
        self.ax.set_xlim(xmin, xmax)
        self.log(3, ('X-axis limits are [ %f, %f], scale: %s' %
                     (xmin, xmax, scale)))
        return

    def setup_yaxis(self, arg_list):
        "Set scale and limits of y-axis by type"
        ymin = self.ymin
        ymax = self.ymax
        scale = 'linear'

        if self.yaxis_type == 'logf':
            # Handle frequency on the Y-axis
            ymin = self.fmin
            ymax = self.fmax
            scale = 'log'
            if arg_list.nologf:
                scale = 'linear'
            if arg_list.fmin:
                ymin = float(arg_list.fmin)
            if arg_list.fmax:
                ymax = float(arg_list.fmax)
        elif self.yaxis_type == 'liny' or self.yaxis_type == 'logy':
            # Handle everything but frequency on Y-axis
            if self.yaxis_type == 'liny':
                scale = 'linear'
                if arg_list.logy:
                    scale = 'log'
            elif self.yaxis_type == 'logy':
                scale = 'log'
                if arg_list.nology:
                    scale = 'linear'
            if arg_list.ymin:
                ymin = float(arg_list.ymin)
            if arg_list.ymax:
                ymax = float(arg_list.ymax)
        else:
            raise AssertionError('Y-axis type is unknown')

        # modify axis
        self.ax.set_yscale(scale)
        self.ax.set_ylim(ymin, ymax)
        self.log(3, ('Y-axis limits are [ %f, %f], scale: %s' %
                     (ymin, ymax, scale)))
        return

    def setup_iaxis(self, arg_list):
        """ set the limits and scale of the colorbar (intensity axis)
        :param arg_list: global arguments
        :return: none
        """

        if self.iaxis_type not in [None, 'lin1', 'log1']:
            raise AssertionError('Unknown intensity axis scale')
        return

    def annotate_save_plot(self, arg_list):
        """After the derived class generated a plot
        object finish the process"""
        from astropy.time import Time
        from gwpy.plotter.tex import label_to_latex

        self.ax = self.plot.gca()
        # set up axes
        self.setup_xaxis(arg_list)
        self.setup_yaxis(arg_list)
        self.setup_iaxis(arg_list)

        if self.is_image():
            if arg_list.nocolorbar:
                self.plot.add_colorbar(visible=False)
            else:
                self.plot.add_colorbar(label=self.get_color_label())
        else:
            self.plot.add_colorbar(visible=False)

        # image plots don't have legends
        if not self.is_image():
            leg = self.ax.legend(prop={'size': 10})
            # if only one series is plotted hide legend
            if self.n_datasets == 1 and leg:
                try:
                    leg.remove()
                except NotImplementedError:
                    leg.set_visible(False)

        # add titles
        title = ''
        if arg_list.title:
            for t in arg_list.title:
                if len(title) > 0:
                    title += "\n"
                title += t
        # info on the processing
        start = self.start_list[0]
        startGPS = Time(start, format='gps')
        timeStr = "%s - %10d (%ds)" % (startGPS.iso, start, self.dur)

        # list the different sample rates available in all time series
        fs_set = set()

        for idx in range(0, len(self.timeseries)):
            fs = self.timeseries[idx].sample_rate
            fs_set.add(fs)

        fs_str = ''
        for fs in fs_set:
            if len(fs_str) > 0:
                fs_str += ', '
            fs_str += '(%s)' % fs

        if self.is_freq_plot:
            spec = r'%s, Fs=%s, secpfft=%.1f (bw=%.3f), overlap=%.2f' %  \
                    (timeStr, fs_str, self.secpfft, 1/self.secpfft,
                     self.overlap)
        else:
            xdur = self.xmax - self.xmin
            spec = r'Fs=%s, duration: %.1f' % (fs_str, xdur)
        spec += ", " + self.filter
        if len(title) > 0:
            title += "\n"
        title += spec

        title = label_to_latex(title)
        self.plot.set_title(title, fontsize=12)
        self.log(3, ('Title is: %s' % title))

        if arg_list.xlabel:
            xlabel = label_to_latex(arg_list.xlabel)
        else:
            xlabel = self.get_xlabel()
        if xlabel:
            self.plot.set_xlabel(xlabel)
            self.log(3, ('X-axis label is: %s' % xlabel))

        if arg_list.ylabel:
            ylabel = label_to_latex(arg_list.ylabel)
        else:
            ylabel = self.get_ylabel(arg_list)

        if ylabel:
            self.plot.set_ylabel(ylabel)
            self.log(3, ('Y-axis label is: %s' % ylabel))

        if not arg_list.nogrid:
            self.ax.grid(b=True, which='major', color='k', linestyle='solid')
            self.ax.grid(b=True, which='minor', color='0.06',
                         linestyle='dotted')

        # info on the channel
        if arg_list.suptitle:
            sup_title = arg_list.suptitle
        else:
            sup_title = self.get_sup_title()
        sup_title = label_to_latex(sup_title)
        self.plot.suptitle(sup_title, fontsize=18)

        self.log(3, ('Super title is: %s' % sup_title))
        self.show_plot_info()

        # change the label for GPS time so Josh is happy
        if self.ax.get_xscale() == 'auto-gps':
            xscale = self.ax.xaxis._scale
            epoch = xscale.get_epoch()
            unit = xscale.get_unit_name()
            utc = re.sub('\.0+', '',
                         Time(epoch, format='gps', scale='utc').iso)
            self.plot.set_xlabel('Time (%s) from %s (%s)' % (unit, utc, epoch))
            self.ax.xaxis._set_scale(unit, epoch=epoch)

        # if they specified an output file write it
        # save the figure. Note type depends on extension of
        # output filename (png, jpg, pdf)
        if arg_list.out:
            out_file = arg_list.out
        else:
            out_file = "./gwpy.png"

        self.log(3, ('xinch: %.2f, yinch: %.2f, dpi: %d' %
                     (self.xinch, self.yinch, self.dpi)))

        self.plot.savefig(out_file, edgecolor='white',
                          figsize=[self.xinch, self.yinch], dpi=self.dpi)
        self.log(3, ('wrote %s' % arg_list.out))

        return

# -----The one that does all the work
    def makePlot(self, args):
        """Make the plot, all actions are generally the same at this level"""
        if args.silent:
            self.verbose = 0
        else:
            self.verbose = args.verbose

        self.log(3, ('Verbosity level: %d' % self.verbose))

        if self.verbose > 2:
            print 'Arguments:'
            for key, value in args.__dict__.iteritems():
                print '%s = %s' % (key, value)

        self.getTimeSeries(args)

        self.config_plot(args)

        # this one is in the derived class
        self.gen_plot(args)

        self.annotate_save_plot(args)

        self.is_interactive = False
        if args.interactive:
            self.log(3, 'Interactive manipulation of '
                        'image should be available.')
            self.plot.show()
            self.is_interactive = True



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Plotting event rate for specific column bins

I would like to study the rate at which event triggers are generated by the
`ExcessPower` gravitational-wave burst detection algorithm, over a small
stretch of data, binned by various thresholds on signal-to-noise ratio (SNR).

The data from which these events were generated are a simulation of Gaussian noise
with the Advanced LIGO design spectrum, and so don't actually contain any real
gravitational waves, but will help tune the algorithm to improve detection of
future, real signals.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__currentmodule__ = 'gwpy.table.lsctables'


# First, we import the `SnglBurstTable` and read the events
from gwpy.table.lsctables import SnglBurstTable
events = SnglBurstTable.read('../../gwpy/tests/data/H1-LDAS_STRAIN-968654552-10.xml.gz')

# Now we can use the :math:`~SnglBurstTable.binned_event_rates` method to calculated the event rate in a number of bins of SNR.
rates = events.binned_event_rates(1, 'snr', [2, 3, 5, 8], operator='>=',
                                  start=968654552, end=968654562)
# .. note::
#    The list `[2, 3, 5, 8]` and operator `>=` specifies SNR tresholds of 
#    2, 3, 5, and 8.

# Finally, we can make a plot:
plot = rates.plot(label='name')
plot.set_xlim(968654552, 968654562)
plot.set_ylabel('Event rate [Hz]')
plot.set_title('LIGO Hanford Observatory event rate for GW100916')
plot.add_legend()
plot.show()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides LIGO.ORG authenticated HTML queries
"""

import os
import stat
import urllib2
import cookielib
import warnings
import tempfile
import getpass
import socket

from glue.auth.saml import HTTPNegotiateAuthHandler

from .. import version

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version

COOKIE_JAR = os.path.join(tempfile.gettempdir(), getpass.getuser())
LIGO_LOGIN_URL = 'login.ligo.org'


def request(url, debug=False, timeout=None):
    """Request the given URL using LIGO.ORG SAML authentication.

    This requires an active Kerberos ticket for the user, to get one:

    .. code-block:: bash

       kinit albert.einstein@LIGO.ORG

    Parameters
    ----------
    url : `str`
        URL path for request
    debug : `bool`, optional
        Query in verbose debugging mode, default `False`
    """
    # set debug to 1 to see all HTTP(s) traffic
    debug = int(debug)

    # need an instance of HTTPS handler to do HTTPS
    httpshandler = urllib2.HTTPSHandler(debuglevel=debug)

    # use a cookie jar to store session cookies
    jar = cookielib.LWPCookieJar()

    # if a cookie jar exists open it and read the cookies
    # and make sure it has the right permissions
    if os.path.exists(COOKIE_JAR):
        os.chmod(COOKIE_JAR, stat.S_IRUSR | stat.S_IWUSR)
        # set ignore_discard so that session cookies are preserved
        try:
            jar.load(COOKIE_JAR, ignore_discard=True)
        except cookielib.LoadError as e:
            warnings.warn('cookielib.LoadError caught: %s' % str(e))

    # create a cookie handler from the cookie jar
    cookiehandler = urllib2.HTTPCookieProcessor(jar)
    # need a redirect handler to follow redirects
    redirecthandler = urllib2.HTTPRedirectHandler()

    # need an auth handler that can do negotiation.
    # input parameter is the Kerberos service principal.
    auth_handler = HTTPNegotiateAuthHandler(
        service_principal='HTTP@%s' % LIGO_LOGIN_URL)

    # create the opener.
    opener = urllib2.build_opener(auth_handler, cookiehandler, httpshandler,
                                  redirecthandler)

    # prepare the request object
    req = urllib2.Request(url)

    # use the opener and the request object to make the request.
    if timeout is None:
        timeout = socket._GLOBAL_DEFAULT_TIMEOUT
    response = opener.open(req, timeout=timeout)

    # save the session cookies to a file so that they can
    # be used again without having to authenticate
    jar.save(COOKIE_JAR, ignore_discard=True)

    return response



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Handle TeX formatting for matplotlib output
"""

from __future__ import division

import os

from astropy import units
from astropy.units.format import (latex as ulatex, utils as uutils)

from .. import version

__author__ = "Duncan M. Macleod <duncan.macleod@ligo.org>"
__version__ = version.version

USE_TEX = os.system('which pdflatex > %s 2>&1' % os.devnull) == 0

LATEX_CONTROL_CHARS = ["%", "\\", "_", "~", "&"]

MACROS = [
    r'\def\rtHz{\ensuremath{\sqrt{\mathrm{Hz}}}}',  # \sqrt{Hz} label
]


def float_to_latex(x, format="%.2g"):
    """Convert a floating point number to a latex representation.

    In particular, scientific notation is handled gracefully: e -> 10^

    Example:
    @code
    >>> float_to_latex(2000)
    '2\times 10^{3}'
    @endcode

    @returns a string in latex mathmode
    """
    base_str = format % x
    if "e" not in base_str:
        return base_str
    mantissa, exponent = base_str.split("e")
    if float(mantissa).is_integer():
        mantissa = int(float(mantissa))
    exponent = exponent.lstrip("0+")
    if exponent.startswith('-0'):
        exponent = '-' + exponent[2:]
    if float(mantissa) == 1.0:
        return r"10^{%s}" % exponent
    else:
        return r"%s\!\!\times\!\!10^{%s}" % (mantissa, exponent)


def label_to_latex(text):
    """Convert an abitrary string of text into a latex-passable
    representation.
    """
    if text is None:
        return ''
    for ch in LATEX_CONTROL_CHARS:
        text = text.replace(ch, "\\%s" % ch)
    return text


def unit_to_latex(unit):
    if isinstance(unit, units.NamedUnit):
        s = label_to_latex(unit.name)
    elif isinstance(unit, units.CompositeUnit):
        if unit.scale != 1:
            s = float_to_latex(unit.scale) + r'\ '
        else:
            s = ''
        if len(unit.bases):
            positives, negatives = uutils.get_grouped_by_powers(
                unit.bases, unit.powers)
            if len(negatives) == 1:
                negatives = format_unit_list(negatives)
                positives = positives and format_unit_list(positives) or 1
                s += r'{0}/{1}'.format(positives, negatives)
            elif len(negatives):
                if len(positives):
                    positives = format_unit_list(positives)
                else:
                    positives = ''
                negatives = format_unit_list(negatives, negative=True)
                s += r'{0}\,{1}'.format(positives, negatives)
            else:
                positives = format_unit_list(positives)
                s += positives
    return r'$\mathrm{{{0}}}$'.format(s)


def format_unit_list(unitlist, negative=False):
    out = []
    texformatter = ulatex.Latex()
    for base, power in unitlist:
        if power == 1 and not negative:
            out.append(texformatter._get_unit_name(base))
        elif power == 0.5 and not negative:
            out.append('\sqrt{{{0}}}'.format(label_to_latex(base.name)))
        elif power != 1 and 1/power == int(1/power):
            out.append('{0}^{{1/{1}}}'.format(
                label_to_latex(base.name), int(1/power)))
        elif negative:
            out.append('{0}^{{-{1}}}'.format(
                label_to_latex(base.name), power))
        else:
            out.append('{0}^{{{1}}}'.format(
                label_to_latex(base.name), power))
    return r'\,'.join(out)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read a `Series` from AN ASCII file

These files should be in two-column x,y format
"""

from numpy import (savetxt, loadtxt)

from astropy.io.registry import (register_reader,
                                 register_writer,
                                 register_identifier)

from ..data import Series
from .. import version
from .utils import identify_factory

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = version.version


def read_ascii(filepath, _obj=Series, xcol=0, ycol=1, delimiter=None,
               **kwargs):
    """Read a `Series` from an ASCII file
    """
    # get specific args for loadtxt
    loadargs = {'unpack': True, 'usecols': [xcol, ycol]}
    for kwarg in ['dtype', 'comments', 'delimiter', 'converters', 'skiprows']:
        if kwarg in kwargs:
            loadargs[kwarg] = kwargs.pop(kwarg)
    # read data, format and return
    x, y = loadtxt(filepath, delimiter=delimiter, **loadargs)
    return _obj(y, xindex=x, **kwargs)


def write_ascii(series, fobj, fmt='%.18e', delimiter=' ', newline='\n',
                header='', footer='', comments='# '):
    """Write a `Series` to a file in ASCII format

    Parameters
    ----------
    series : :class:`~gwpy.data.Series`
        data series to write
    fobj : `str`, `file`
        file object, or path to file, to write to

    See also
    --------
    numpy.savetxt : for documentation of keyword arguments
    """
    x = series.xindex.value
    y = series.value
    return savetxt(fobj, zip(x, y), fmt=fmt, delimiter=delimiter,
                   newline=newline, header=header, footer=footer,
                   comments=comments)


formats = {'txt': None,
           'csv': ','}


def ascii_io_factory(obj, delimiter=None):
    def _read(filepath, **kwargs):
        kwargs.setdefault('delimiter', delimiter)
        return read_ascii(filepath, _obj=obj, **kwargs)

    def _write(series, filepath, **kwargs):
        kwargs.setdefault('delimiter', delimiter or ' ')
        return write_ascii(series, filepath, **kwargs)
    return _read, _write


def register_ascii(obj):
    """Register ASCII I/O methods for given type obj

    This factory method registers 'txt' and 'csv' I/O formats with
    a reader, writer, and auto-identifier
    """
    for form, delim in formats.iteritems():
        read_, write_ = ascii_io_factory(obj, delim)
        register_identifier(form, obj, identify_factory(form, '%s.gz' % form))
        register_writer(form, obj, write_)
        register_reader(form, obj, read_)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) Joseph Areeda (2015)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.
#

""" Time Series plots
"""
from cliproduct import CliProduct


class TimeSeries(CliProduct):

    def get_action(self):
        """Return the string used as "action" on command line."""
        return 'timeseries'

    def init_cli(self, parser):
        """Set up the argument list for this product"""
        self.arg_chan1(parser)
        self.arg_ax_linx(parser)
        self.arg_ax_liny(parser)
        self.arg_plot(parser)
        return

    def get_ylabel(self, args):
        """Text for y-axis label"""
        return 'Counts'

    def get_title(self):
        """Start of default super title, first channel is appended to it"""
        return 'Time series: '

    def gen_plot(self, args):
        """Generate the plot from time series and arguments"""
        self.max_size = 16384. * 6400.  # that works on my mac
        self.yscale_factor = 1.0

        from gwpy.plotter.tex import label_to_latex
        from numpy import min as npmin
        from numpy import max as npmax

        if self.timeseries[0].size <= self.max_size:
            self.plot = self.timeseries[0].plot()
        else:
            self.plot = self.timeseries[0].plot(linestyle='None', marker='.')
        self.ymin = self.timeseries[0].min().value
        self.ymax = self.timeseries[0].max().value
        self.xmin = self.timeseries[0].times.value.min()
        self.xmax = self.timeseries[0].times.value.max()

        if len(self.timeseries) > 1:
            for idx in range(1, len(self.timeseries)):
                chname = self.timeseries[idx].channel.name
                lbl = label_to_latex(chname)
                if self.timeseries[idx].size <= self.max_size:
                    self.plot.add_timeseries(self.timeseries[idx], label=lbl)
                else:
                    self.plot.add_timeseries(self.timeseries[idx], label=lbl,
                                             linestyle='None', marker='.')
                self.ymin = min(self.ymin, self.timeseries[idx].min().value)
                self.ymax = max(self.ymax, self.timeseries[idx].max().value)
                self.xmin = min(self.xmin,
                                self.timeseries[idx].times.value.min())
                self.xmax = max(self.xmax,
                                self.timeseries[idx].times.value.max())
        # if they chose to set the range of the x-axis find the range of y
        strt = self.xmin
        stop = self.xmax
        # a bit weird but global ymax will be >= any value in
        # the range same for ymin
        new_ymin = self.ymax
        new_ymax = self.ymin

        if args.xmin:
            strt = float(args.xmin)
        if args.xmax:
            stop = float(args.xmax)
        if strt != self.xmin or stop != self.xmax:
            for idx in range(0, len(self.timeseries)):
                x0 = self.timeseries[idx].x0.value
                dt = self.timeseries[idx].dt.value
                if strt < 1e8:
                    strt += x0
                if stop < 1e8:
                    stop += x0
                b = int(max(0, (strt - x0) / dt))

                e = int(min(self.xmax, (stop - x0) / dt))

                if e >= self.timeseries[idx].size:
                    e = self.timeseries[idx].size - 1
                new_ymin = min(new_ymin, npmin(self.timeseries[idx].value[b:e]))
                new_ymax = max(new_ymax, npmax(self.timeseries[idx].value[b:e]))
            self.ymin = new_ymin
            self.ymax = new_ymax
        if self.yscale_factor > 1:
            self.log(2, ('Scaling y-limits, original: %f, %f)' %
                         (self.ymin, self.ymax)))
            yrange = self.ymax - self.ymin
            mid = (self.ymax + self.ymin) / 2.
            self.ymax = mid + yrange / (2 * self.yscale_factor)
            self.ymin = mid - yrange / (2 * self.yscale_factor)
            self.log(2, ('Scaling y-limits, new: %f, %f)' %
                         (self.ymin, self.ymax)))
        return
